function [avLL, warnPercent] = compAvLL_robust(tl, w0, ...
    testdata, margin, setbackMargin, setbackLength)
% Computes average log likelihoods of OOM tl, w0 on
% sequences given as rows in the r x s matrix testdata.
% The algorithm uses the "robust" sequence running algorithm with
% control parameters margin, setbackMargin, setbackLength, as
% described in Appendix J in H. Jaeger et al,
% "Learning Observable Operator Models via the ES Algorithm", in "New
% Directions in Statistical Signal Processing: from Systems to Brains"
% (eds. Simon Haykin, Jose Principe, Terry Sejnowski, John McWhirter), MIT Press 2005.
% If the OOM is valid (never produces negative probs), the robust algorithm
% yields the same results as a generic likelihood computation algorithm would do. If there
% are negative probabilities, the returned loglikelihood is slightly lower than 
% optimally could be the case because the negative probs are "blurred
% away", which amounts to a loss of model accuracy. 
% 
% input arguments:
% tl: array sized dim x dim x alphabetsize, containing the observable
%     operators
% w0: starting state
% testdata: test sequences contained as rows in matrix of size
%           (number of test sequences) x (length of each test sequence)
% margin: small positive real number, typical settings 0.001 to 0.01
% setbackMargin: not quite so small nonnegative real number, typical
%                 choices 0 to 0.5
% setbackLength: nonnegative integer, standard choices are 2 or 3
%
% Outputs:
% avLL: average log likelihood on test sequences, computing with the robust
%       algorithm compLL_robust
% warnPercent: percent of occasions where running the OOM encountered an
%              unrecoverable negative probability. Should be zero! If not, the OOM is
%              highly instable and should not be used.
%
% V 1.0, May 2005
% (c) Herbert Jaeger
%
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.

alphabetsize = size(tl,3);
stringLength = length(testdata(1,:));
nrOfStrings = length(testdata(:,1));
LLmodelList = zeros(1,nrOfStrings); 
warnsCollector = 0;

for i = 1:nrOfStrings
    seq = testdata(i,:);       
    [LLmodel, warns, setbacks, actCounter, changeCounter] = ...
    compLL_robust(tl, w0, seq, margin, setbackMargin, setbackLength); 
    warnsCollector = warnsCollector + warns;
    LLmodelList(1,i) = LLmodel;    
end

avLL = mean(LLmodelList);
warnPercent = 100 * warnsCollector / (stringLength * nrOfStrings);