function [loglik, warns, setbacks, actCounter, changeCounter] = ...
    compLL_robust(tl, w0, seq, margin, setbackMargin, setbackLength)
% computes log likelihood of an OOM given by tl and w0 on a sequence seq. 
% Inputs margin, setbackMargin, setbackLength are control parameters as
% described in Appendix J in H. Jaeger et al,
% "Learning Observable Operator Models via the ES Algorithm", in "New
% Directions in Statistical Signal Processing: from Systems to Brains"
% (eds. Simon Haykin, Jose Principe, Terry Sejnowski, John McWhirter), MIT Press 2005.
% 
% input arguments:
% tl: array sized dim x dim x alphabetsize, containing the observable
%     operators
% w0: starting state
% seq: symbol sequence, row vector of integers
% margin: small positive real number, typical settings 0.001 to 0.01
% setback margin: not quite so small nonnegative real number, typical
%                 choices 0 to 0.5
% setbackLength: nonnegative integer, standard choices are 2 or 3
%
% Outputs:
% loglik: log likelihood of sequence under model tl, w0, using the
%         "blurred" robustifying algorithm which leads to a little decrease in the
%         loglik value, trading loglik quality against robustness.
% setbacks: number of occasions where the state went unruly to an extent
%           that the setback operation (restarting from earlier state) was carried
%           out.
% actCounter: number of occasions where some P(a|w_n) was negative and
%             corrective action was taken, that is, the negative prob mass was
%             subtracted from the probs  P(a'|w_n) symbols a' with positive prob, etc.,
%             as described in  the Haykin book chapter, Appendix J.
% changeCounter: number of occasions where the actually observed symbol a
%                had a negative probability under the OOM (and which is compensated away).
% warns: counter of how many times a negative prob could not be corrected
%        away. 
% Note: nonzero values of actCounter imply an invalid OOM.
% Note: high values of actCounter or changeCounter are no reason for concern.
% Note: a nonzero value of warns is indicative of very instable model;
%       the resulting loglik output is very likely incorrect (too high); 
%       results (and the OOM) should not be used!!
%
% V 1.0, May 2005
% (c) Herbert Jaeger
%
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.

dim = size(tl,1); alphabetsize = size(tl,3);
l = length(seq);

tlSqueezed = zeros(alphabetsize , dim);
for a = 1:alphabetsize
    tlSqueezed(a,:) = sum(tl(:,:,a)); 
end

tlSums = sum(tlSqueezed');

w = w0; loglik = 0; actCounter = 0; 
setbacks = 0; warns = 0; changeCounter = 0;

for n = 1:l
    % probs is the distribution vector over symbols for predicted next step
    probs = tlSqueezed * w; 
    
    % deltaSum is the total negative prob mass of predicted 1-step
    % distribution. otherProbSum is sum of nonnegative values of predicted 
    % 1-step distribution. actedFlag jumps to 1 if there is at least one
    % negative prob
    deltaSum = 0; otherProbSum = 0;  actedFlag = 0;
    
    for a = 1:alphabetsize        
        if probs(a) < 0 
            deltaSum = deltaSum + margin - probs(a);
            actedFlag = 1;
        else
            otherProbSum = otherProbSum + probs(a);
        end
    end    
    
    % carry out setback if current state is ill-behaved
    if deltaSum > setbackMargin & n <= setbackLength + 1 % a setback is indicated but it is 
        % too early in the sequence for setback
        w = w0;              
        setbacks  = setbacks + 1;
        probs = tlSqueezed * w; 
        deltaSum = 0; otherProbSum = 0; actedFlag = 0;
        for a = 1:alphabetsize
            if probs(a) < 0 
                deltaSum = deltaSum + margin - probs(a);
                actedFlag = 1;
            else
                otherProbSum = otherProbSum + probs(a);
            end
        end  
    elseif deltaSum > setbackMargin % situation after initial setbackLength observations
        % set back state
        setbacks  = setbacks + 1;
        w = w0;
        word = seq(n - setbackLength:n-1);
        for i = 1:setbackLength
            w = tl(:,:,word(i)) * w;
            w = w/sum(w);            
        end
        % recompute deltaSum etc
        probs = tlSqueezed * w; 
        deltaSum = 0; otherProbSum = 0; actedFlag = 0;
        for a = 1:alphabetsize
            if probs(a) < 0 
                deltaSum = deltaSum + margin - probs(a);
                actedFlag = 1;
            else
                otherProbSum = otherProbSum + probs(a);
            end
        end          
    end    
    
    if actedFlag
        actCounter = actCounter + 1;
    end    
    
    remainingProbMass = otherProbSum - deltaSum; % prob mass that unchanged ops have to add up to
    normalizingFactor = remainingProbMass / otherProbSum; 
    
    a = seq(n);
    if tlSums(a) == 0
        loglik = NaN; warns = NaN;  setbacks = NaN;  actCounter = NaN;  changeCounter = NaN; 
        disp('Warning: compLL_robust was run with model that had zero operator for some symbol');
        return
    end
    
    if  probs(a) <= 0 
        changeCounter = changeCounter + 1;
        w = tl(:,:,a) * w;        
        w = w / sum(w);
        prob = margin;        
    else
        w = tl(:,:,a) * w;
        w = w / sum(w);
        prob = probs(a) * normalizingFactor;
    end
    
    if prob > 0
        logs = log(prob);
    else 
        warns = warns + 1; 
        logs = -1;
    end  
    loglik = loglik + logs;    
end
  
