% A sample script for generating training and testing data and learning and
% testing them with the "Poor Man's" and the suffix-tree based ES learning 
% algorithm, as well as
% comparing the OOM models with HMM models trained by Baum-Welch.
% Training data come from a MOOM, that is, an OOM whose operator matrices are
% all non-negative (for more about MOOMs, see the NIPS 2005 paper by
% Jaeger/Zhao/Kolling, downloadable from my publications page).
%
% Check out the headers of the function files below for more detailed
% documentation.
%
% Version 1.0, June 6, 2005
% (c) Herbert Jaeger
%
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.

%%%% generate a random,  interpretable generator MOOM
gendim = 4; alphabetsize = 3; Exponent = 10; figNr = 1;
[tlInt, w0Int, C, charEvLength] = genRandMOOM(gendim, alphabetsize, Exponent, figNr);
    

%%%% compute training string
trainDataLength = 1000;
trainData = genSeq(tlInt, w0Int, trainDataLength); % training sequence is row vector
% compute test strings, sorted row-wise into testseqs which has dimension
% nrOfStrings x stringlength
stringLength = 10000; nrOfStrings = 1; 
testseqs = zeros(nrOfStrings, stringLength); 
for n = 1:nrOfStrings
    testseqs(n,:) = genSeq(tlInt, w0Int, stringLength);
end

% set dim for models
moddim = 4;

%%%%%%%%%%%%%%%%% Learn Poor Man's ES model %%%%%%%%%%%%%%%%%

% Setting parameters for ES/poor man's learning. For a synopsis inspect header of
% file learnOOM_PoorMan.
evLength = 2; dimTol = 0.0001; setbackLength = 2; margin = 0.01; 
setbackMargin = 0.3; iter = 5; mkInter = 1; 
verbose = 2; modelSelectionMode = 2; figNrs = [2,3]; printLegend = 0; 
             
% run the poor man's learning algorithm (see header of learnOOM_PoorMan for
% explanation of outputs)
[tlLearnt, w0Learnt, Cfwd, CondPL,   LLTrainPL, LLTestPL, WeightPL] =  ...
    learnOOM_PoorMan(moddim, alphabetsize, trainData, testseqs, evLength, dimTol, setbackLength,   ...
    margin, setbackMargin, iter, mkInter, verbose, modelSelectionMode, figNrs, printLegend);

% plot the generator fingerprint in the same interpretable format as 
% fingerprint plots produced by learnOOM_PoorMan
figure(4);   clf; 
plotOOMSequence_interpretable(tlInt, w0Int, Cfwd, evLength, trainData(1:500)); 
title('generator, same C as Poor Man Model'); 

disp('compute training / testing LL of Poor Man trained OOM:');
% compute average log likelihood of model on training Data
[LLtrain, warnPercentTrain] = ...
    compAvLL_robust(tlLearnt, w0Learnt, trainData, margin, setbackMargin, setbackLength);
disp(sprintf('Average train LL = %g   percentage of irrecoverable negative probs = %g',...
    LLtrain, warnPercentTrain));
% compute average log likelihood of model on testsequences
[LLtest, warnPercentTest] = ...
    compAvLL_robust(tlLearnt, w0Learnt, testseqs, margin, setbackMargin, setbackLength);
disp(sprintf('Average test LL = %g   percentage of irrecoverable negative probs = %g',...
    LLtest, warnPercentTest));

%%%%%%%%%%%%%%%%% Learning an OOM with the suffix-tree ES algorithm %%%%%%%%%%%%%%%%%

evLength = 2; dimTol = 0.0001;
setbackLength = 2; margin = 0.001; setbackMargin = 0.3; 
weedPercent = 1;  minLeafNr = 2; maxIndLength = 20; 
iter = 5; mkInter = 1; recomputeST = 1; 
WienerHopf = 0; verbose = 2; modelSelectionMode = 2; figNrs = [5,6]; printLegend = 0;


[tlLearnt_ST, w0Learnt_ST, Cfwd_ST, LLTrainBPL_ST, WeightPL_ST, CondPL_ST, LLTrainPL_ST, LLTestPL_ST] =  ...
    learnOOM_SuffixTree(moddim, alphabetsize, trainData,  testseqs, evLength, dimTol, setbackLength,   ...
    margin, setbackMargin, weedPercent, minLeafNr, maxIndLength, iter, ...
    mkInter, recomputeST, WienerHopf, verbose, modelSelectionMode, figNrs, printLegend);

figure(7);   clf; 
plotOOMSequence_interpretable(tlInt, w0Int, Cfwd_ST, evLength, trainData(1:500)); 
title('generator, same C as suffix tree Model'); 

disp('compute training / testing LL of suffix-tree trained OOM:');
% compute average log likelihood of model on training Data
[LLtrain_ST, warnPercentTrain_ST] = ...
    compAvLL_robust(tlLearnt_ST, w0Learnt_ST, trainData, margin, setbackMargin, setbackLength);
disp(sprintf('Average train LL = %g   percentage of irrecoverable negative probs = %g',...
    LLtrain_ST, warnPercentTrain_ST));
% compute average log likelihood of model on testsequences
[LLtest_ST, warnPercentTest_ST] = ...
    compAvLL_robust(tlLearnt_ST, w0Learnt_ST, testseqs, margin, setbackMargin, setbackLength);
disp(sprintf('Average test LL = %g   percentage of irrecoverable negative probs = %g',...
    LLtest_ST, warnPercentTest_ST));

%%%%%%%%%%%%%%%%% HMM training %%%%%%%%%%%%%%%%%%%%%%%%%%
% For comparison, a HMM is trained with EM, using the public domain implementation
% written by Kevin Murphy, obtainable at http://www.ai.mit.edu/~murphyk/Software/HMM/hmm.html
% This code calls helpers from the public domain Netlab toolbox written by 
%  Ian T Nabney, obtainable at http://www.ncrg.aston.ac.uk/netlab/ .
% For convenience, I copied the few files that are needed for our purposes
% into the OOM package. Kevin's implementation of EM for HMMs had a small bug
% that led to an unneccessary huge consummation of memory; this is fixed
% in the files dhmm_em_Herbert and fwdback_Herbert.m that I supply here
% instead of the original files dhmm_em.m and fwdback.m . 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('********* HMM/Baum Welch training **********');
% train HMM
EMdim = moddim; maxEMiterations = 50; EMthresh = 5e-5;
startMarkovMat = mk_stochastic(rand(EMdim,EMdim));
startObsMat = mk_stochastic(rand(EMdim,alphabetsize));
startW0Hmm = normalise(rand(EMdim,1));
[LL, w0HMMest, markovMatEst, obsmatEst] = ...
    dhmm_em_Herbert(trainData, startW0Hmm, ...
    startMarkovMat, startObsMat, 'max_iter', maxEMiterations, 'thresh', EMthresh, 'verbose',1);
% transform to equivalent OOM
diagmatsEst = zeros(EMdim, EMdim, alphabetsize);
for a = 1:alphabetsize
    diagmatsEst(:,:,a) = diag(obsmatEst(:,a));
end            
tlHMM = HMM2OOM(markovMatEst, diagmatsEst);
w0HMM = compW0(tlHMM); 
% compute LL on test data
[LLtestHMM, warnPercentTestHMM] = ...
    compAvLL_robust(tlHMM, w0HMM, testseqs, margin, setbackMargin, setbackLength);
disp(sprintf('HMM Loglik on training data = %g   on test data = %g', ...
     LL(end), LLtestHMM));
% create a plot similar (and directly comparable) to the ones plotted for
% the OOM models
figure(8); clf; 
plotOOMSequence_interpretable(tlHMM, w0HMM, Cfwd, evLength, trainData(1:500)); 
title('HMM model, same C as suffix tree model');




