% A sample script for training an OOM using the suffix tree based
% variant of the learning algorithm, on a training string derived from
% Mark Twain's short story "The 1 Mio Pound Note". 
%
% Check out the headers of the function files below for more detailed
% documentation.
%
% Version 1.0, May 11, 2005
% (c) Herbert Jaeger
%
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.



% the main suffix tree constructs are declared global because they are
% typically used repeatedly throughout a modelling session 
% and need to be computed only once for a given training data set. 
global cst slink  fTable  nodeNr nodePerms; % globals of suffix tree

% load the train and test data. They were obtained by sorting the original
% short story sentence-wise into two text strings (for training and
% testing), then deleting all special characters except the blank, setting
% all letters to smallcap, and finally coding letters by integers 
% (a -> 1,..., z  -> 26, blank -> 27)
load mpound_train.txt -ascii;
trainData = mpound_train(1:end,1)';
load mpound_test.txt -ascii;
testseqs = mpound_test(1:end,1)';

%%%% Setting parameters for OOM learning. For a synopsis, inspect header of
%%%% file learnOOM_SuffixTree. Note that the setting "verbose = 2" yields a
%%%% maximum of diagnostic output -- and needs a maximum of runtime.
%%%% Setting "verbose = 0" will very much accelerate the affair. 
dim = 10; alphabetsize = 27; evLength = 1; dimTol = 0.00001;
setbackLength = 2; margin = 0.001; setbackMargin = 0.3; 
weedPercent = 1;  minLeafNr = 10; maxIndLength = 50; 
iter = 5; mkInter = 1; recomputeST = 1; 
WienerHopf = 0; verbose = 2; modelSelectionMode = 2; figNrs = [1,2]; printLegend = 0;


[tlLearnt, w0Learnt, Cfwd, LLTrainBPL, WeightPL, CondPL, LLTrainPL, LLTestPL] =  ...
    learnOOM_SuffixTree(dim, alphabetsize, trainData,  testseqs, evLength, dimTol, setbackLength,   ...
    margin, setbackMargin, weedPercent, minLeafNr, maxIndLength, iter, ...
    mkInter, recomputeST, WienerHopf, verbose, modelSelectionMode, figNrs, printLegend);


% compute average log likelihood of model on training Data
disp('compute training / testing LL of suffix-tree trained OOM:');
[LLtest, warnPercentTrain] = ...
    compAvLL_robust(tlLearnt, w0Learnt, trainData, margin, setbackMargin, setbackLength);
disp(sprintf('Train LL = %g   percentage of irrecoverable negative probs = %g',...
    LLtest, warnPercentTrain));
% compute average log likelihood of model on testsequences
[LLtest, warnPercentTest] = ...
    compAvLL_robust(tlLearnt, w0Learnt, testseqs, margin, setbackMargin, setbackLength);
disp(sprintf('Test LL = %g   percentage of irrecoverable negative probs = %g',...
    LLtest, warnPercentTest));

%%%%%%%%%%%%%%%%% HMM training %%%%%%%%%%%%%%%%%%%%%%%%%%
% For comparison, a HMM is trained with EM, using the public domain implementation
% written by Kevin Murphy, obtainable at http://www.ai.mit.edu/~murphyk/Software/HMM/hmm.html
% This code calls helpers from the public domain Netlab toolbox written by 
% Ian T Nabney, obtainable at http://www.ncrg.aston.ac.uk/netlab/ .
% For convenience, I copied the few files that are needed for our purposes
% into the OOM package. Kevin's implementation of EM for HMMs had a small bug
% that led to an unneccessary huge consummation of memory; this is fixed
% in the files dhmm_em_Herbert and fwdback_Herbert.m that I supply here
% instead of the original files dhmm_em.m and fwdback.m . 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('********* learning a HMM/Baum-Welch model ***********');
% train HMM
EMdim = dim; maxEMiterations = 50; EMthresh = 1e-5;
startMarkovMat = mk_stochastic(rand(EMdim,EMdim));
startObsMat = mk_stochastic(rand(EMdim,alphabetsize));
startW0Hmm = normalise(rand(EMdim,1));
[LL, w0HMMest, markovMatEst, obsmatEst] = ...
    dhmm_em_Herbert(trainData, startW0Hmm, ...
    startMarkovMat, startObsMat, 'max_iter', maxEMiterations, 'thresh', EMthresh, 'verbose',1);
% transform to equivalent OOM
diagmatsEst = zeros(EMdim, EMdim, alphabetsize);
for a = 1:alphabetsize
    diagmatsEst(:,:,a) = diag(obsmatEst(:,a));
end            
tlHMM = HMM2OOM(markovMatEst, diagmatsEst);
w0HMM = compW0(tlHMM); 
% compute LL on test data
[LLtestHMM, warnPercentTestHMM] = ...
    compAvLL_robust(tlHMM, w0HMM, testseqs, margin, setbackMargin, setbackLength);
disp(sprintf('HMM Loglik on training data = %g   on test data = %g', ...
     LL(end), LLtestHMM));
% create a plot similar (and directly comparable) to the ones plotted for
% the OOM models
figure(3); clf; 
plotOOMSequence_interpretable(tlHMM, w0HMM, Cfwd, evLength, trainData(1:500)); 
title('HMM model');



