function [chisquare, confLevel] = dimCheck(tl, w0, trainSetSize, lss)
% Check whether the dimension of the OOM (tl, w0) is justified given that
% the model was estimated from trainSetSize many data points. 
%
% Inputs:
%
% tl: array sized dim x dim x alphabetsize, containing the observable
%     operators
% w0: starting state
% trainSetSize: number of data points used for training (tl, w0)
% lss: length of the state sequence which is generated to assess the
%      invariant cone of the OOM. The larger, the more precisely are the
%      least important principal components determined - at the cost of
%      computing more states.
%
% Outputs:
%
% chisquare: the test statistics computed for this case
% confLevel: a probability value, which can be interpreted as "with a 
%            probability of at least confLevel, the current model
%            capacity is not too large, reducing the dimension would 
%            not improve the test performance". 
%
% Created July 16, 2007, H. Jaeger
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.

dim = size(tl,1); alphabetsize = size(tl,3);

% compute a characteristic event length suitable for this alphabetsize and
% dim
charEvLength = 1;
while alphabetsize^charEvLength < dim
    charEvLength = charEvLength + 1;
end
% compute degrees of freedom
df = (alphabetsize - 1) * (dim - 1)^2;  
% compute the "bigV" matrix from the current trained model
[bigVtest, PiMattest, stateMattest] = bigVfromModel(tl, w0, charEvLength);
% derive model with dim reduced by 1            
[tl_dimRed, w0_dimRed, delSV_MassRatio] = reduceDim(tl, w0, lss, 1, 0, 0);
% compute the "bigV" matrix from the dim-reduced model
[bigVref, PiMatref, stateMatref] = bigVfromModel(tl_dimRed, w0_dimRed, charEvLength);
% normalize the two "bigV" matrices to mass l
bigVtest = trainSetSize * bigVtest / sum(sum(bigVtest));
bigVref = trainSetSize * bigVref / sum(sum(bigVref));
% compute chisquare test statistic
chisquare = sum(sum(((bigVtest - bigVref).^2) ./ bigVref));
% compute confidence level
confLevel = gammainc(chisquare/2, df/2);
