function [C, w0] = findCharEventsFromBigV(bigV, dim, l)
% Find good characteristic events from a "big-sized" matrix bigV =  
%  (\hat{P}(\bar{a}_j \bar{a}_i)), where the \bar{a}_i are
% all words of length evLength. The algorithm is explained in Appendix I in H. Jaeger et al,
% "Learning Observable Operator Models via the ES Algorithm", in "New
% Directions in Statistical Signal Processing: from Systems to Brains"
% (eds. Simon Haykin, Jose Principe, Terry Sejnowski, John McWhirter), MIT Press 2005.  
%
% input arguments:
%
% bigV: matrix of size \kappa x \kappa, where 
%       \kappa = alphabetsize^(length of characteristic events)
% dim: dimension of OOM for which the char. events are computed
% l: length of training data from which bigV was estimated. Is needed to
%    compute a bias term. If set to 0, infinite data length (that is,
%    correct bigV) is assumed and no bias added
% outputs:
% 
% C: a binary matrix of size dim x \kappa, the characteristic event
%    characterizer. See Haykin book chapter Section 11 for an explanation
% w0 is rowsum vector of C * bigV, the starting state for the interpretable
%    OMM resulting from using C
%
% V 1.0, May 2005
% (c) Herbert Jaeger
%
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.

kappa = size(bigV,1);
C = zeros(dim, kappa);
bigV_rowsums = sum(bigV'); % contains the relative frequencies of the \bar{a}_i

bigVjoined = full(bigV); 
% add bias to bigVjoined
if l > 0
    for i = 1:kappa
        if bigV_rowsums(i) ~= 0
            bigVjoined(i,:) = (1 - 1/l) * bigVjoined(i,:) + bigV_rowsums(i) * ones(1,kappa) / (kappa * l);
        end
    end
end

for i = 1:kappa
    assemble(i).indList = [i];
end


remainingRows = sum(sign(bigV_rowsums)); % number on non-zero prob \bar{a}_i
if remainingRows < dim
    disp('Error: bigV has not enough non-zero prob rows for targetted moddim in findCharEventsFromBigV. Reduce dim or increase size of bigV'); 
end
remainingProb = 1;

% normalize rows of bigV to unit norm
bigVn = bigVjoined;
for i = 1:kappa
    if bigV_rowsums(i) ~= 0
        bigVn(i,:) = bigVn(i,:) / norm(bigVn(i,:));
    end
end

% make all those \bar{a}_i singleton char events whose prob is larger than
% 1/dim
finishedCharEvents = 0;
for i = 1:kappa
    if bigV_rowsums(i) > remainingProb / (dim - finishedCharEvents)
        finishedCharEvents = finishedCharEvents + 1;
        remainingRows = remainingRows - 1;
        C(finishedCharEvents, i) = 1;
        bigVn(i,:) = zeros(1,kappa);
        bigVjoined(i,:) = zeros(1,kappa);
        remainingProb = remainingProb - bigV_rowsums(i);         
    end
end

cM = bigVn * bigVn';
cM = cM - diag(diag(cM));

% while there remain charevents to join, join.
while remainingRows > dim - finishedCharEvents & finishedCharEvents < dim         
    % compute correlation matrix of rows of bigVn and set diagonal to zero
    
    [maxColVals, maxRowInds] = max(cM);
    [maxRowVal, maxColInd] = max(maxColVals);
    maxRowInd = maxRowInds(maxColInd); % now (maxRowInd, maxColInd) is pair of rows to join
    joinInds = sort([maxRowInd, maxColInd]);
    newIndListraw = [assemble(joinInds(1)).indList assemble(joinInds(2)).indList];
    newIndListraw = sort(newIndListraw);
    newIndList = [newIndListraw(1)];
    for n = 2:length(newIndListraw)
        if newIndList(end) < newIndListraw(n)
            newIndList = [newIndList newIndListraw(n)];
        end
    end
    assemble(joinInds(1)).indList = newIndList;  
    
    % join rows in bigVjoined
    bigVjoined = [bigVjoined(1:joinInds(1)-1, :); ...
            bigVjoined(joinInds(1), :) + bigVjoined(joinInds(2), :);...
            bigVjoined(joinInds(1)+ 1 : joinInds(2)- 1, :); ...
            zeros(1,kappa); ...
            bigVjoined(joinInds(2)+1:end, :)    ];
    remainingRows = remainingRows - 1;
    % compute things from old bigVn needed below for updating cM
    oldBigVnRow1 = bigVn(joinInds(1),:); oldBigVn = bigVn;
    row1 = - oldBigVnRow1 + bigVjoined(joinInds(1), :) / norm(bigVjoined(joinInds(1), :)); 
    row2 = - bigVn(joinInds(2),:);
    col1 = bigVn * row1'; col2 = bigVn * row2';
    p12 = row1 * row2'; p21 = row2 * row1'; 
    % adjust bigVn
    bigVn(joinInds(2),:) = zeros(1,kappa);
    bigVn(joinInds(1),:) = bigVjoined(joinInds(1),:);
    bigVn(joinInds(1),:) = bigVn(joinInds(1),:) / norm(bigVn(joinInds(1),:)); 
    
    % if sum of joined event is large enough, make it the next charevent
    if sum(bigVjoined(joinInds(1),:)) > remainingProb / (dim - finishedCharEvents)        
        finishedCharEvents = finishedCharEvents + 1;
        C(finishedCharEvents, assemble(joinInds(1)).indList) = ...
            ones(1,length(assemble(joinInds(1)).indList));
        remainingProb = remainingProb - sum(bigVjoined(joinInds(1),:));
        remainingRows = remainingRows - 1;
        bigVjoined(joinInds(1),:) = zeros(1,kappa);        
        row1 = - oldBigVnRow1; 
        p12 = row1 * row2'; p21 = row2 * row1'; 
        col1 = oldBigVn * row1'; col2 = oldBigVn * row2';
        bigVn(joinInds(1),:) = zeros(1,kappa);  
    end  
    % update cM   
    cM(:, joinInds(1)) = cM(:, joinInds(1)) + col1;
    cM(:, joinInds(2)) = cM(:, joinInds(2)) + col2;
    cM(joinInds(1),:) = cM(joinInds(1),:) + col1'; 
    cM(joinInds(2),:) = cM(joinInds(2),:) + col2';
    cM(joinInds(1), joinInds(2)) = cM(joinInds(1), joinInds(2)) + p21;
    cM(joinInds(2), joinInds(1)) = cM(joinInds(2), joinInds(1)) + p12;
    cM(joinInds(1), joinInds(1)) = 0;
    cM(joinInds(2), joinInds(2)) = 0;
end

% now there remain as many unfinished charevents as there are nonzero rows
% in bigVjoined. They are distributed over the remaining charevents
for i = 1:kappa
    if sum(bigVjoined(i,:)) ~= 0 
        finishedCharEvents = finishedCharEvents + 1;
        C(finishedCharEvents, assemble(i).indList) = ones(1,length(assemble(i).indList));       
    end
end

% finally, where bigV has zero rows, put a 1 into the last C row
for i = 1:kappa
    if bigV_rowsums(i) == 0 
        C(end, i) = 1;   
    end
end

w0 = (sum((C * full(bigV))'))';

