function [tlInt, w0Int,C,charEvLength] = genRandHMMOOM(dim, alphabetsize, Mexponent, Oexponent, figNr)
% Creates a random state-emitting HMM and transforms it to an interpretable OOM. 
%
% Input arguments: 
% dim: number of hidden HMM states, also the OOM dimension
% alphabetsize: number of observables
% Mexponent: exponent of value distribution of Markov matrix entries. When
% set to 1, the entries of the Markov transition matrix are sampled from
% a uniform distribution. Higher values lead to sample distributions that
% are increasingly skewed to the right, with a few high entries in the
% transition matrix and many low ones. 
% Oexponent: exponent of value distribution of of HMM emission  matrix. Has
% similar effect as Mexponent, but for the HMM emission probabilities
% figNr: if not 0, a fingerprint plot of the generated OOM is displayed in
% figure with this figNr
% 
% Output arguments:
% tlInt: array of tau_a operators of an interpretable OOM derived
% from the HMM, size dim x dim x alphabetsize
% w0Int: starting state of the OOM
% C: characteristic event characterizer used to make OOM interpretable.
%    For an explanation, check out Section 11 in H. Jaeger et al,
%    "Learning Observable Operator Models via the ES Algorithm", in "New
%    Directions in Statistical Signal Processing: from Systems to Brains"
%    (eds. Simon Haykin, Jose Principe, Terry Sejnowski, John McWhirter), MIT Press 2005. 
% charEvLength: length of characteristic events used to make OOM interpretable.
%
% Version 1.0, May 11, 2005
% Copyright Herbert Jaeger
%
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.

continueTrying = 1; 
% we enclose the entire procedure in a retry loop because sometimes it
% happens that the randomly created HMM describes a process with process
% dimension that is numerically less than the full dimension, in which case
% the transformation to an OOM fails due to ill-conditioned matrix inverse.
while continueTrying  
    try,
        % construct Markov matrix
        markovMat = zeros(dim,dim);
        for i = 1:dim
            for j = 1:dim            
                markovMat(i,j) = rand^Mexponent;            
            end
        end    
        %normalize to rowsums 1
        for r = 1:dim
            markovMat(r,:) = markovMat(r,:) / sum(markovMat(r,:));
        end
        
        % construct observation matrix
        obsmat = zeros(dim,alphabetsize);
        for i = 1:dim
            for j = 1:alphabetsize            
                obsmat(i,j) = rand^Oexponent;            
            end
        end
        % normalize to rowsums 1
        for r = 1:dim
            if sum(obsmat(r,:)) ~= 0
                obsmat(r,:) = obsmat(r,:) / sum(obsmat(r,:));
            else
                obsmat(r,1) = 1;
            end
        end
                
        diagMats = zeros(dim, dim, alphabetsize);
        for a = 1:alphabetsize
            diagMats(:,:,a) = diag(obsmat(:,a));
        end
        
        tl = HMM2OOM(markovMat, diagMats);
        w0 = compW0(tl);
        [tlInt, w0Int, rho, C, charEvLength] = makeInterpretable_Auto(tl,w0);        
        continueTrying = 0;
    catch,
        continueTrying = 1;
    end
end

if figNr ~= 0 & dim < 3
    disp('Warning: cannot produce state plot of OOM with dim < 3 when calling genRandHMMOOM');
end
if figNr ~= 0 & dim >= 3
    figure(figNr);clf;
    plotOOMFingerprint(tlInt, w0Int, 500);
    title('OOM state plot');
end
