function [tlInt, w0Int,C,charEvLength] = genRandMOOM(dim, alphabetsize, Exponent, figNr)
% Creates a random transition-emitting HMM (equivalent to non-negative OOM) 
% and transforms it to an interpretable OOM. 
%
% Input arguments: 
% dim: number of hidden TE-HMM states, also the OOM dimension
% alphabetsize: number of observables
% Exponent: exponent of value distribution of tau operator matrix entries. When
% set to 1, the entries of the matrices are sampled from
% a uniform distribution. Higher values lead to sample distributions that
% are increasingly skewed to the right, with a few high entries in the
% transition matrix and many low ones. 
% figNr: if not 0, a fingerprint plot of the generated OOM is displayed in
% figure with this figNr
% 
% Output arguments:
% tlInt: array of tau_a operators of an interpretable OOM derived
% from the HMM, size dim x dim x alphabetsize
% w0Int: starting state of the OOM
% C: characteristic event characterizer used to make OOM interpretable.
%    For an explanation, check out Section 11 in H. Jaeger et al,
%    "Learning Observable Operator Models via the ES Algorithm", in "New
%    Directions in Statistical Signal Processing: from Systems to Brains"
%    (eds. Simon Haykin, Jose Principe, Terry Sejnowski, John McWhirter), MIT Press 2005. 
% charEvLength: length of characteristic events used to make OOM interpretable.
%
% Version 1.0, May 11, 2005
% Copyright Herbert Jaeger
%
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.

continueTrying = 1; 
% we enclose the entire procedure in a retry loop because sometimes it
% happens that the randomly created TE-HMM describes a process with process
% dimension that is numerically less than the full dimension, in which case
% the transformation to an OOM fails due to ill-conditioned matrix inverse.
while continueTrying  
    try,        
        tl = rand(dim, dim, alphabetsize).^Exponent;
        mu = sum(tl,3);
        musums = sum(mu);
        %normalize 
        for d = 1:dim
            for a = 1:alphabetsize
                tl(:,d,a) = tl(:,d,a) / musums(d);
            end            
        end        
        w0 = compW0(tl);
        [tlInt, w0Int, rho, C, charEvLength] = makeInterpretable_Auto(tl,w0);        
        continueTrying = 0;
    catch,
        continueTrying = 1;
    end
end

if figNr ~= 0 & dim < 3
    disp('Warning: cannot produce state plot of OOM with dim < 3 when calling genRandMOOM');
end
if figNr ~= 0 & dim >= 3
    figure(figNr);clf;
    plotOOMFingerprint(tlInt, w0Int, 500);
    title('OOM state plot');
end
