function [tlLearnt, w0Learnt, Cfwd, CondPL, LLTrainPL, LLTestPL, WeightPL] =  ...
    learnOOM_PoorMan(dim, alphabetsize, trainData, testseqs, evLength, dimTol, setbackLength,   ...
    margin, setbackMargin, iter, mkInter, verbose, modelSelectionMode, figNrs, printLegend)
% Learn a stationary, discrete OOM from a (long, single) training string, using the
% "poor man's" ES algorithm described Section 11 of H. Jaeger et al,
% "Learning Observable Operator Models via the ES Algorithm", in "New
% Directions in Statistical Signal Processing: from Systems to Brains"
% (eds. Simon Haykin, Jose Principe, Terry Sejnowski, John McWhirter), MIT Press 2005.
%
%%%%%%%%%%%% input arguments %%%%%%%%%%%%
% dim:
% Model dimension
%
% alphabetsize:
% Number of symbols for which model is trained. Must be at least as 
% big as the maximal entry in trainData, and is
% preferably identical to max(trainData). If
% alphabetsize > [nr. of different symbols in trainData], a model is learnt with
% zero operators for the symbol indices not occurring in
% training data.
%
% trainData: 
% a row vector representing a symbol string by integers.
%
% testseqs: 
% an n x L sized matrix containing n >= 1 test sequences, each of length L, in its rows
%
% evLength:
% Length of characteristic sequences used for initialization model estimation.
%
% dimTol:
% tolerance level for checking whether computed model has effective dim
% that is as large as requested dimension dim. dimTol is threshold for
% singular value size for matrix (\hat(P)(\bar{a}_i \bar{b}_j). The larger
% dimTol, the earlier is a model rejected. A value of 0 never rejects a
% model. 
%
%%%%%%  The next three arguments tune the heuristic algorithms for OOM sequence 
%%%%%%  generation that are called
%%%%%%  inside the learning algorithm. This heuristic algorithm copes robustly
%%%%%%  with the negative probabilities that arise with invalid OOMs, which are
%%%%%%  typically obtained through learning. For an explanation of this heuristic
%%%%%%  algorithm and the three parameters, check out Appendix J in H. Jaeger et al,
%%%%%%  "Learning Observable Operator Models via the ES Algorithm", in "New
%%%%%%  Directions in Statistical Signal Processing: from Systems to Brains"
%%%%%%  (eds. Simon Haykin, Jose Principe, Terry Sejnowski, John McWhirter), MIT Press 2005. 
%
% setbackLength:
% If invalid states are encountered, how many states are stepped back and restarted from
%
% margin:
% The positive probability by which negative or zero probabilities are
% replaced. Recommended range 0.001 to 0.01
%
% setbackMargin: 
% Threshold for triggering the setback routine. If the next-step pseudo-probability of any 
% symbol falls below -setbackMargin, the set-backing is triggered.
% Recommended range 0.1 to 0.5.
%
% iter:
% Number of iterations (excluding the initial model estimation, which will however be
% covered in the first element of the output lists CondPL, LLTrainPL, LLTestPL, WeightPL)
% 
% mkInter:
% Set this flag to 1 (suggested default) if the OOMs computed in each iteration 
% are to be made interpretable. Set to 0 if not.
% The value of this flag has in theory no effect on the
% quality of models, except that in critical cases, when the
% models become very ill-conditioned, making them
% interpretable may improve numerical stability. The main benefit
% however is to make the models obtained in the iterations
% visually comparable. So, if you run the learning algorithm
% in verbose = 2 mode, producing "fingerprint" output, you should set
% this to 1. 
%
% verbose: 
% Determines amount (and cost) of diagnostic output. Possible settings
% 0,1,2.
% If verbose is 0, no diagnostics are printed. The (cheaply computed) outputs 
% LLTrainBPL and WeightPL are computed but the (expensive) outputs  
% CondPL, LLTrainPL, LLTestPL are set to []. 
% If verbose is 1, moderate diagnostics is printed but the #PL
% outputs still are treated as with verbose = 0.
% If verbose is 2, detailed diagnostics (expensive to compute) 
% is printed, figs are plotted, and all ouput arguments are
% assigned.
%
% modelSelectionMode:
% Set to 1: model obtained in last iteration is returned.
% Set to 2: model from iteration with best likelihood on reverse 
% training string is returned.
% 
% figNrs:
% a pair [f1, f2] of integers indicating the figure numbers into which the
% fingerprint plots and the learning dynamics diagnostics are to be plotted
%
% printLegend:
% Set to 1 when after learning run a legend to the diagnostic output should 
% be printed (for first-time users)
%
%
% ************* outputs *************
% tlLearnt: array of size dim x dim x alphabetsize containing the
%           operators of the learng model
% w0Learnt: initial state of the learnt model
% Cfwd: the characteristic event characterizer used in the initial model
%       estimation.
% CondPL: a row vector of length iter + 1 containing the condition numbers
%         of the argument matrices C * \underline(V) found throughout the
%         iterations. For explanation see the Haykin book chapter, Section
%         11. This is only computed when verbose = 2 or = 1; else it is set to NaN's.
% LLTrainPL: a row vector of length iter + 1 containing the training
%            loglikelihoods found throughout the iterations.  This is only 
%            computed when verbose = 2; else it is set to NaN's.
% LLTestPL: a row vector of length iter + 1 containing the testing 
%            loglikelihoods (averaged over test strings) found throughout 
%            the iterations.  This is only 
%            computed when verbose = 2; else it is set to NaN's.
% WeightPL: a row vector of length iter + 1 containing the summed absolute weights
%           of the model operator matrices found throughout 
%            the iterations. 
%
% V 1.0, May 2005
% (c) Herbert Jaeger 
%
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.

if verbose > 0
    disp(' *************  executing learnOOM_PoorMan...  ************* ');
end

f1 = figNrs(1); f2 = figNrs(2);
tl = zeros(dim, dim, alphabetsize);
w0 = zeros(dim,1);
tlLearnt = NaN * tl;
w0Learnt = NaN * w0;
kappa = alphabetsize ^ evLength;
if kappa < dim
    disp('Error: evLength is too small for model dimension');
end

% setting the stage for plotting
statePlotFlag = 0;
if verbose == 2 & (dim < 3 | iter > 48)
    disp('Warning: with more than 48 iterations or dim < 3, no state plots will be produced');
end
if iter <= 48 & verbose == 2 & dim >= 3 % set the panel grid for the fingerprint figure
    statePlotFlag = 1;    
    if iter <= 3
        plotGrid = 2;
    elseif iter <= 8
        plotGrid = 3;
    elseif iter <= 15
        plotGrid = 4;
    elseif iter <= 24
        plotGrid = 5;
    elseif iter <= 35
        plotGrid = 6;
    else
        plotGrid = 7;
    end
    figure(f1); clf;    
end

% init the diagnostic output lists
WeightPL = NaN * zeros(1,iter + 1  );
CondPL = NaN * zeros(1,iter + 1  ); 
LLTrainPL = NaN * zeros(1,iter + 1  );
LLTestPL =  NaN * zeros(1,iter + 1  );

l = length(trainData);  

% print training data histogram
if verbose > 0
    histogram = zeros(1,alphabetsize);
    for i = 1:l
        histogram(1,trainData(1,i)) = histogram(1,trainData(1,i)) + 1;
    end   
    disp(sprintf('trainData histogram %s', mat2str(histogram)));    
end

% compute the kappa x kappa sized counting statistics matrix 
% \underline{V} = (\hat{P}(\bar{a}_j \bar{b}_i))
% and the corresponding 
% \underline{W}_a = (\hat{P}(\bar{a}_j a \bar{b}_i))
[bigV, bigWas] = bigVWfromData(evLength, trainData, alphabetsize);

% compute characteristic event characterizer for initial model estimation
[Cfwd, w0fwd] = findCharEventsFromBigV(bigV, dim, l);

% re-weigh bigV, bigWas into BigVNorm, bigWasNorm according to the remark after 
% "two embellishments", Section11, from the Haykin Book chapter.
% First reconstitute raw counting matrices
bigVraw = bigV * (length(trainData) -  2*evLength);
bigWasraw = bigWas;
for a = 1:alphabetsize
    bigWasraw(a).W = bigWasraw(a).W * (length(trainData) -  2*evLength);
end
% Second, reweigh
bigVnorm = bigVraw;
bigWasnorm = bigWasraw;
colfacs = sqrt(sum(bigVraw));
% colfacs = (sum(bigVraw));
for j = 1:kappa
    if colfacs(j) ~= 0
        bigVnorm(:,j) = bigVnorm(:,j) / colfacs(j);
        for a = 1:alphabetsize
            bigWasnorm(a).W(:,j) = bigWasnorm(a).W(:,j) / colfacs(j);
        end
    end
end

% compute initial model with the "classical" OOM learning algorithm
V = Cfwd * bigVnorm;
pinvV = pinv(V);
for a = 1:alphabetsize
    tl(:,:,a) = Cfwd * bigWasnorm(a).W  * pinvV;
end
w0 = compW0(tl);

tlOld = tl;

% exit with NaN results if model has effective dim that is lower than dim
[bigVtest, PiMattest, stateMattest] = bigVfromModel(tl,w0,evLength);
effDim = rank(bigVtest, dimTol);
if effDim < dim
    disp('Warning: effective initial model dim smaller than requested dim. Exit from learnOOM_PoorMan');
    return
end

if mkInter % make model interpretable w.r.t. characteristic event characterizer used
        % for initial model
    [tl, rho]  = makeInterpretable(tl,Cfwd, evLength);
    w0 = rho * w0;        
end


% if needed for diagnostics or model selection, compute forward state
% sequence and model loglik
if verbose > 0 | modelSelectionMode == 2
    [stateseqFwd, loglik, warns, setbacks, actCounter, changeCounter] = ...
      compStateSeq_robust(tl, w0 , trainData, margin, setbackMargin, setbackLength); 
end

% if best loglik model is to be returned, assign current model to
% output and its loglik to loglikBest
if modelSelectionMode == 2    
        loglikBest = loglik;
        tlLearnt = tl; w0Learnt = w0;    
end

% a cheap diagnostic item, always computed
modelWeight = sum(sum(sum(abs(tl),3)));
WeightPL(1) = modelWeight;

% print/collect diagnostics
if verbose == 1
    condition = cond(V); 
    CondPL(:,1) = condition;
    disp(sprintf('it %g: LL %0.1f  W %0.1f  cond %0.1f  negs %g   setb %g   act %g  changes %g',...
        0, loglik,  modelWeight, condition, warns, setbacks, actCounter, changeCounter ));
elseif verbose == 2
    [LLtest, warnPercentTest] = ...
        compAvLL_robust(tl, w0, testseqs, margin, setbackMargin, setbackLength);    
    condition = cond(V);
    LLTrainPL(1) = loglik;
    LLTestPL(1) = LLtest;
    CondPL(:,1) = condition;         
    disp(sprintf('it %g: LLtr %0.1f  LLts %g  W %0.1f  cond %0.1f   negs %g   setb %g   act %g  changes %g ',...
        0, loglik,  LLtest, modelWeight,  condition, warns, setbacks, actCounter, changeCounter));    
end

% plot "fingerprint" of initial model
if  statePlotFlag        
    figure(f1);
    subplot(plotGrid, plotGrid, 1);
    plotOOMStateseq(stateseqFwd(2:501,:), trainData(1:500), alphabetsize);  
    title('initial model');
end 

tlOld = tl;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%% ES iterations %%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for iteration = 1:iter   
    % compute reverse OOM
    
    [tlBack, w0Back] = reverseOOM(tl); 
%     tlBack(:,:,1)
%     w0Back
    % compute new characterizer C, using current reverse model
    C = zeros(dim, kappa);
    for i = 1:kappa
        % generate i-th sequence seq in alphabetic enumeration
        seq = zeros(1,evLength);
        j = i-1;
        for k = evLength : - 1 : 1        
            ind =  floor((j)/alphabetsize^(k-1));
            j = j - ind * alphabetsize^(k-1);        
            seq(1,evLength-k+1) = ind + 1;
        end
        w = w0Back;
        for n = evLength:-1:1
            w = tlBack(:,:,seq(n)) * w; 
            sumw = sum(w);
            if sumw ~= 0
                w = w / sumw;
            end
        end
        C(:,i) = w;
    end
    
    % recompute model    
    V = C * bigVnorm ;
    pinvV = pinv(V);    
    for a = 1:alphabetsize
        tl(:,:,a) = C * bigWasnorm(a).W  * pinvV;
    end
    
    w0 = compW0(tl);
    
     % exit with NaN results if model has effective dim that is lower than dim
    [bigVtest, PiMattest, stateMattest] = bigVfromModel(tl,w0,evLength);
    effDim = rank(bigVtest, dimTol);
    if effDim < dim
        disp(sprintf(...
            'Warning: effective model dim in iteration %g smaller than requested dim. Exit from learnOOM_PoorMan',...
            iteration));
        tlLearnt = NaN * zeros(dim, dim, alphabetsize);
        w0Learnt = NaN * zeros(dim,1);
        WeightPL = NaN * zeros(1,iter + 1  );
        CondPL = NaN * zeros(1,iter + 1  ); 
        LLTrainPL = NaN * zeros(1,iter + 1  );
        LLTestPL =  NaN * zeros(1,iter + 1  );
        return
    end
    
    if mkInter % make model interpretable w.r.t. characteristic event characterizer if desired
        [tl, rho]  = makeInterpretable(tl,Cfwd, evLength);
        w0 = rho * w0;        
    end
    

    if verbose > 0 | modelSelectionMode == 2
        [stateseqFwd, loglik, warns, setbacks, actCounter, changeCounter] = ...
          compStateSeq_robust(tl, w0 , trainData, margin, setbackMargin, setbackLength); 
    end
    if modelSelectionMode == 2
        if loglik > loglikBest
            loglikBest = loglik;
            tlLearnt = tl; w0Learnt = w0;
        end
    end
    
    modelWeight = sum(sum(sum(abs(tl),3)));
    WeightPL(iteration + 1) = modelWeight;
    
    if verbose == 1  
        condition = cond(V);
        CondPL(iteration + 1) = condition; 
        disp(sprintf('it %g: LL %0.1f  W %0.1f  cond %0.1f  negs %g   setb %g   act %g  changes %g',...
            iteration, loglik,  modelWeight, condition, warns, setbacks, actCounter, changeCounter ));
    elseif verbose == 2
        [LLtest, warnPercentTest] = ...
            compAvLL_robust(tl, w0, testseqs, margin, setbackMargin, setbackLength);        
        condition = cond(V);
        LLTrainPL(iteration + 1) = loglik;        
        LLTestPL(iteration + 1) = LLtest;        
        CondPL(iteration + 1) = condition; 
        disp(sprintf('it %g: LLtr %0.1f  LLts %g  W %0.1f  cond %0.1f   negs %g   setb %g   act %g  changes %g ',...
            iteration, loglik,  LLtest, modelWeight,  condition, warns, setbacks, actCounter, changeCounter));        
    end    
    
    if  statePlotFlag  % plot fingerprint      
        figure(f1);
        subplot(plotGrid, plotGrid,iteration + 1);
        plotOOMStateseq(stateseqFwd(2:501,:), trainData(1:500), alphabetsize);   
        if iteration == 1
            title('Poor Man models...');
        end
    end  
    
    tlOld = tl;
end

if printLegend & verbose == 1
    disp('********* Explanation of diagnostic categories *********');
    disp('it: iteration');
    disp('LL: log likelihood of model on train data');
    disp('W: summed absolute matrix entries of (all) operator matrices');
    disp('cond: condition number of argument matrix V');
    disp('negs: Nr of unrecovered negative probs encountered when computing state sequence');
    disp('       Should be 0, else model is dubiously unstable');
    disp('setb: Nr of emergency recovery setback operations during computing state sequence');
    disp('act: Nr of "blurred" probabilities used during computing state sequence');
    disp('changes: Nr of recovered negative probs encountered when computing state sequence');
end
if printLegend & verbose == 2
    disp('********* Explanation of diagnostic categories *********');
    disp('it: iteration');
    disp('LLtr: log likelihood of model on train data');
    disp('LLts: average log likelihood of model on test data');
    disp('W: summed absolute matrix entries of (all) operator matrices');
    disp('cond: condition number of argument matrix V');
    disp('negs: Nr of unrecovered negative probs encountered when computing state sequence');
    disp('       Should be 0, else model is dubiously unstable');
    disp('setb: Nr of emergency recovery setback operations during computing state sequence');
    disp('act: Nr of "blurred" probabilities used during computing state sequence');
    disp('changes: Nr of recovered negative probs encountered when computing state sequence');
end

if modelSelectionMode == 1
    tlLearnt = tl;
    w0Learnt = w0;
end

if verbose == 2
    figure(f2);clf;    
    subplot(4,1,1);
    plot(LLTestPL'); 
    title('LLtest'); 
    subplot(4,1,2);
    plot(LLTrainPL); 
    title('LLtrain');    
    subplot(4,1,3);
    plot(WeightPL');
    title('Weights');
    subplot(4,1,4);
    plot(CondPL); 
    title('condition numbers');    
end

if verbose > 0
    disp(' *************  finished learnOOM_PoorMan  ************* ');
end
