function [tlLearnt, w0Learnt, Cfwd, LLTrainBPL, WeightPL, CondPL, LLTrainPL, LLTestPL] =  ...
    learnOOM_SuffixTree(dim, alphabetsize, trainData, testseqs, evLength, dimTol, setbackLength,   ...
    margin, setbackMargin, weedPercent, minLeafNr, maxIndLength, iter, ...
    mkInter, recomputeST, WienerHopf, verbose, modelSelectionMode, figNrs, printLegend)
% Learn a stationary, discrete OOM from a (long, single) training string, using the
% suffix-tree based ES algorithm described Sections 12 and 13 of H. Jaeger et al,
% "Learning Observable Operator Models via the ES Algorithm", in "New
% Directions in Statistical Signal Processing: from Systems to Brains"
% (eds. Simon Haykin, Jose Principe, Terry Sejnowski, John McWhirter), MIT Press 2005.
%
%
% Version 1.0, May 11, 2005
% (c) Herbert Jaeger
% Rev. 1, July 7, 2007, H. Jaeger (corrected error message)
%
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.
%
%%%%%%%%%%%%% Input arguments %%%%%%%%%%%%%
% dim:
% Model dimension
%
% alphabetsize:
% Number of symbols for which model is trained. Must be at least as 
% big as the maximal entry in trainData, and is
% preferably identical to max(trainData). If
% alphabetsize > [nr. of different symbols in trainData], a model is learnt with
% zero operators for the symbol indices not occurring in
% training data.
%
% trainData: 
% a row vector representing a symbol string by integers.
%
% testseqs: 
% an n x L sized matrix containing n >= 1 test sequences, each of length L, in its rows
%
% evLength:
% Length of characteristic sequences used for initialization model estimation.
%
% dimTol:
% tolerance level for checking whether computed model has effective dim
% that is as large as requested dimension dim. dimTol is threshold for
% singular value size for matrix (\hat(P)(\bar{a}_i \bar{b}_j). The larger
% dimTol, the earlier is a model rejected. A value of 0 never rejects a
% model. 
%
%%%%%%  The next three arguments tune the heuristic algorithms for OOM sequence 
%%%%%%  generation that are called
%%%%%%  inside the learning algorithm. This heuristic algorithm copes robustly
%%%%%%  with the negative probabilities that arise with invalid OOMs, which are
%%%%%%  typically obtained through learning. For an explanation of this heuristic
%%%%%%  algorithm and the three parameters, check out Appendix J in H. Jaeger et al,
%%%%%%  "Learning Observable Operator Models via the ES Algorithm", in "New
%%%%%%  Directions in Statistical Signal Processing: from Systems to Brains"
%%%%%%  (eds. Simon Haykin, Jose Principe, Terry Sejnowski, John McWhirter), MIT Press 2005. 
%
% setbackLength:
% If invalid states are encountered, how many states are stepped back and restarted from
%
% margin:
% The positive probability by which negative or zero probabilities are
% replaced. Recommended range 0.001 to 0.01
%
% setbackMargin: 
% Threshold for triggering the setback routine. If the next-step pseudo-probability of any 
% symbol falls below -setbackMargin, the set-backing is triggered.
% Recommended range 0.1 to 0.5.
%
% weedPercent:
% Percentage of states of reverse state sequences which are replaced by the 
% "certainly harmless" initial state, before the reverse
% state sequence is filtered into the suffix tree in each
% iteration. States with maximal summed absolute
% components are replaced; the rationale being that such
% states are very likely artifact states due to invalid
% OOM behaviour. Thus this is a kind of crude outlier
% "de-fusing". Typical value range 0 to 1; with 0 chosen
% if the learning task is easy (model dimension low
% relative to data size).
%
% minLeafNr:
% Minimal number of occurrences of a substring before its suffix tree 
% node is used for model estimation. Typical range 1 to 5.
% The larger this is chosen, the faster the algorithm
% because the effective size of the to-be-inverted V matrix
% shrinks.
%
% maxIndLength:
% Maximum length of indicative sequences chosen from the suffix tree. 
% Thus only counting statistics of indicative sequences
% up to this length effectively enter the model
% estimation. Typical choice 10, the more data
% available, the larger this may be reasonably chosen.
% The smaller this is chosen, the faster the algorithm
% because the effective size of the to-be-inverted V
% matrix shrinks.      
%
% iter:
% Number of iterations (including the initial model estimation)
% 
% mkInter:
% Set this flag to 1 (suggested default) if the OOMs computed in each iteration 
% are to be made interpretable. Set to 0 if not.
% The value of this flag has in theory no effect on the
% quality of models, except that in critical cases, when the
% models become very ill-conditioned, making them
% interpretable may improve numerical stability. The main benefit
% however is to make the models obtained in the iterations
% visually comparable. So, if you run the learning algorithm
% in verbose = 2 mode, producing graphical output, you should set
% this to 1. 
%
% recomputeST:
% 0-1-flag indicating whether the suffix tree is to be recomputed. While 
% a fixed dataset is being used for training different OOMs
% (e.g., tuning model dimension or control parameters),
% the suffix tree has not to be recomputed in every
% learning run - so set this to 1 only in the first
% learning run of an experimenting session. (Computing the
% suffix tree costs some time)
%
% WienerHopf:
% 0-1-flag indicating whether the Wiener Hopf equation is to be used 
% for computing the main linear regression of the model
% estimation, or whether the pseudoinverse is to be used.
% The latter is numerically more robust but slightly more
% computationally expensive. A setting of 0 is recommended as default. Only if 
% memory capacity is about to be surpassed or the last bit of speed squeezing
% is desired, one should use WienerHopf = 1.
%
% verbose: 
% Determines amount (and cost) of diagnostic output. Possible settings
% 0,1,2.
% If verbose is 0, no diagnostics are printed. The (cheaply computed) outputs 
% LLTrainBPL and WeightPL are computed but the (expensive) outputs  
% CondPL, LLTrainPL, LLTestPL are set to []. 
% If verbose is 1, moderate diagnostics is printed but the #PL
% outputs still are treated as with verbose = 0.
% If verbose is 2, detailed diagnostics (expensive to compute) 
% is printed, figs are plotted, and all ouput arguments are
% assigned.
%
% modelSelectionMode:
% Set to 1: model obtained in last iteration is returned.
% Set to 2: model from iteration with best likelihood on reverse 
% training string is returned.
% 
% figNrs:
% a pair [f1, f2] of integers indicating the figure numbers into which the
% fingerprint plots and the learning dynamics diagnostics are to be plotted
%
% printLegend:
% Set to 1 when after learning run a legend to the diagnostic output should 
% be printed (for first-time users)
%
%
%%%%%%%%%%%%% Outputs %%%%%%%%%%%%%
% tlLearnt:
% Array of size dim x dim x alphabetsize, containing the observable
% operator matrices of the learnt OOM
%
% w0Learnt:
% The invariant starting state of the learnt model
%
% Cfwd: 
% The characteristic event characterizer used in the initial model estimation.
% If verbose = 2 and mkInter = 1, the plotted fingerprints come from OOMs
% that are interpretable w.r.t. this characterizer.
% 
% LLTrainBPL: 
% List of length iter containing the loglikelihoods of the reverse models
% obtrained during the iterations.
%
% WeightPL: 
% List of length iter  containing the summed absolute weights
% of the model operator matrices found throughout the iterations. 
%
% CondPL: 
% List of length iter containing the condition numbers of the argument
% matrices V_normalized found throughout the
% iterations. For explanation see the Haykin book chapter, Section
% 11. This is only computed when verbose = 2 or = 1; else it is set to [].
%
% LLTrainPL: 
% List of length iter containing the training
% loglikelihoods of the "forward" models found throughout the iterations.  
% This is only computed when verbose = 2; else it is set to [].
% 
% LLTestPL: 
% List of length iter containing the testing 
% loglikelihoods (averaged over test strings) found throughout the iterations.  
% This is only computed when verbose = 2; else it is set to [].


global cst slink fTable nodeNr nodePerms; % the suffix tree globals

if verbose > 0
    disp(' ************* executing learnOOM_SuffixTree ... ************ ');
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%% initializations %%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

f1 = figNrs(1); f2 = figNrs(2);
tl = zeros(dim, dim, alphabetsize);
w0 = zeros(dim,1);
tlLearnt = NaN * tl;
w0Learnt = NaN * w0;
loglikBest = -1e99; % keeps track of the backward logliks of models
kappa = alphabetsize ^ evLength;
if kappa < dim
    disp('Error in learnOOM_SuffixTree: evLength is too small for model dimension');
    return
end

% setting the stage for plotting
statePlotFlag = 0;
if verbose == 2 && (dim < 3 || iter > 49)
    disp('Warning: with more than 49 iterations or dim < 3, no state plots will be produced');
end
if iter <= 49 && verbose == 2 && dim >= 3  % set the panel grid for the fingerprint figure
    statePlotFlag = 1;    
    if iter <= 4
        plotGrid = 2;
    elseif iter <= 9
        plotGrid = 3;
    elseif iter <= 16
        plotGrid = 4;
    elseif iter <= 25
        plotGrid = 5;
    elseif iter <= 36
        plotGrid = 6;
    else
        plotGrid = 7;
    end     
end

% init the diagnostic output lists
LLTrainBPL = NaN * zeros(1,iter ); 
WeightPL = NaN * zeros(1,iter ); 
LLTrainPL = NaN * zeros(1,iter );
LLTestPL = NaN * zeros(1,iter );
CondPL = NaN * zeros(1,iter );  


l = length(trainData);  
trainDataRev = reverseMat(trainData,2); % the reverse training string

% print training data histogram
if verbose > 0
    histogram = zeros(1,alphabetsize);
    for i = 1:l
        histogram(1,trainData(1,i)) = histogram(1,trainData(1,i)) + 1;
    end   
    disp(sprintf('trainData histogram %s', mat2str(histogram)));    
end

% compute the kappa x kappa sized counting statistics matrix (\hat{P}(\bar{a}_j \bar{b}_i))
bigV = bigVfromData(evLength, trainData, alphabetsize);
% run heuristic algorithm for finding good characteristic events, which are
% coded in characteristic event characterizer Cfwd
[Cfwd, w0fwd] = findCharEventsFromBigV(bigV, dim, l);
clear bigV;

% compute a dummy reverse state sequence whose states are 0-1-vectors with
% a single 1 indicating the characteristic event to which the substring
% after the concerned state is assigned
stateseqBack = zeros(l+1,dim);
stateseqBack(end - evLength + 1:end,:) = ones(evLength,1) * w0fwd';
for n = 1:l+1 - evLength
    window = trainData(n:n+evLength - 1);
    charEvInd = charIndex(window,  alphabetsize);
    indVec = zeros(kappa,1);
    indVec(charEvInd) = 1;    
    stateseqBack(n,:) = (Cfwd * indVec)';   
end

% compute the suffix tree and its assorted indexing tables
if recomputeST   
    if verbose > 0
        disp('computing suffix tree... ');
    end
    [cst, slink, childInfo] = sfxtree([trainDataRev 0]); % the suffix tree essentials  
    clear childInfo; % not needed but big
    nodeNr = cst(1,1); % nr of nodes in ST
    [xx, nodePerms] = sort(cst(:,9)); % nodePerms reorders nodes with 
    % increasing distance from leaf
    % level, thus the first indices in
    % nodePerms point to leaf nodes 
    
    % Compute the lookup table of size nodeNr x alphabetsize. 
    % If the n-th suffix tree node corresponds to an indicative sequence
    % \bar{a}, then the (n,b)-th entry of fTable points to the highest suffix tree
    % node whose indicative sequence ends with \bar{a}b. If \bar{a}b does not occur 
    % in the training string, the (n,b)-th entry of fTable points to nodeNr + 1
    fTable = (nodeNr + 1) * ones(nodeNr, alphabetsize); % the lookup table
    fCandidates = zeros(nodeNr, alphabetsize); % a temporary helper table    
    % after indicative event of nodeNr
    for n = nodePerms'  % we visit the ST nodes from leaves upwards    
        if cst(n,9) == 0 % case of seeing a leave
            % find leaf for suffix that is longer by 1 than current leave's
            % suffix. This is the first leaf node above current node in cst
            if cst(n,6) ~= l+1 % only do something for leaves that are not the leaf of the longest suffix
                longerLeafNode = n - 1;
                while cst(longerLeafNode, 3) ~= 0
                    longerLeafNode = longerLeafNode - 1;                    
                end % after this while loop, longerLeafNode points to the desired leaf                
                extensionSymbol = trainDataRev(1,end - cst(n,6) + 1);
                fTable(n,extensionSymbol) = longerLeafNode;
                fCandidates(cst(n,1),extensionSymbol) = longerLeafNode;                
            end
        else % case of non-leaf nodes
            for a = 1:alphabetsize
                if fCandidates(n,a) ~= 0 % If this is 0, the current node's string from root to
                    % node cannot be extended by a.
                    % Now find that node above fCandidates(n,a) whose
                    % root-to-node string has minimal length greater or
                    % equal (the length of the root-to-node string of node n
                    % + 1) [= fitlength]
                    fitlength = cst(n,6) + 1;
                    candNode = fCandidates(n,a);
                    candidateLength = cst(candNode,6);
                    while fitlength <= candidateLength - cst(candNode,11)                    
                        candidateLength = candidateLength - cst(candNode,11);
                        candNode = cst(candNode,1);
                    end % after this while loop, candNode points to desired node
                    if candNode ~= 0 
                        fTable(n,a) = candNode;
                    end
                    if n ~= 1
                        fCandidates(cst(n,1),a) = candNode;
                    end
                end
            end
        end
    end  
    clear fCandidates;
    if verbose > 0
        disp('... suffix tree computed. ');
    end
end

tlOld = zeros(dim, dim, alphabetsize);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%% iterations %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for iteration = 1:iter
    
    % "De-fuse" weedPercent of normalized (that is, w0 is subtracted) 
    % states from stateseqBack that have
    % maximal  summed absolute entries, by replacing them with w0. Rationale:
    % the replaced states are considered as outlier artifacts due to invalid OOM.
    if iteration > 1  &  weedPercent > 0
        stateseqBackAbsVals = sum(abs((stateseqBack - ones(l+1,1) * w0Back')'));
        [absVals, indicesBack] = sort(stateseqBackAbsVals);
        nrWeeds = floor(weedPercent * l / 100);        
        weedIndsBack = indicesBack(end - nrWeeds + 1:end);        
        for i = weedIndsBack        
            stateseqBack(i,:) = w0Back';
        end
    end
    
    % Iteratively sum states associated with leaves upwards through suffix
    % tree. Omit the leaf corresponding to the entire sequence. This will 
    % yield a proto version protoV of the argument matrix V.  This protoV
    % contains (summed) states in its columns, which in turn correspond to
    % nodes of the ST. 
    protoV = zeros(dim,nodeNr);    
    for n = nodePerms' % we go upwards through the ST, starting at the leaf level
        if cst(n,9) == 0 % case of seeing a leave               
            if cst(n,6) ~= l+1 % leave out the node corresponding to complete sequence                
                protoV(:,n) = stateseqBack(l+1 - cst(n,7) + 1,:)';                  
            end
        else
            protoV(:,n) = protoV(:,cst(n,3)); 
            brother = cst(cst(n,3),2); % find index of first brother of first child
            while brother ~= 0
                protoV(:,n) = protoV(:,n) + protoV(:,brother);
                brother = cst(brother,2);
            end
        end
    end    
    % collapse protoV to those columns corresponding to indicative
    % sequences not longer than maxIndLength, and appearing at least
    % minLeafNr times in the string - that is, such indicative sequences
    % that will actually be used.
    useIndices = find(cst(:,10) >= minLeafNr & cst(:,6) <= maxIndLength);
    nrUsed = length(useIndices);
    protoV = protoV(:,useIndices);
    
    % compute the observable operators, either using the Wiener Hopf
    % equation or the pseudoinverse
    if WienerHopf
        % construct a normalized version of protoV where colums sum to 1
        if length(protoV(:,1)) > 1 
            normSums = sum(protoV);
        else
            normSums = protoV; % for 1-dim models
        end
        V_normalized = protoV;
        for n = 1:nrUsed 
            if normSums(1,n) ~= 0
                V_normalized(:,n) = protoV(:,n) / normSums(1,n);
            end
        end
        
        if verbose >= 1  % a diagnostic item, computed only in the more expensive modes      
            condition = cond(V_normalized);        
        end
        
        % compute Wiener Hopf covariance matrix
        covMat = V_normalized * protoV';
        clear protoV;
        covMatInv = inv(covMat) ;
        clear covMat;
        
        % protoW has largely similar structure as protoV. It contains the
        % target (summed state) vectors for the linear regression in its columns. 
        % Iteratively sum states associated with leaves upwards through suffix
        % tree. Unlike for protoV, include the leaf corresponding to the
        % entire sequence. 
        protoW = zeros(dim,nodeNr + 1);  % in last column will contain zero vector        
        for n = nodePerms'
            if cst(n,9) == 0 % case of seeing a leave            
                protoW(:,n) = stateseqBack(l+1 - cst(n,7) + 1,:)';                
            else
                protoW(:,n) = protoW(:,cst(n,3)); % init with summed states of first child                
                brother = cst(cst(n,3),2); % find index of first brother of first child
                while brother ~= 0
                    protoW(:,n) = protoW(:,n) + protoW(:,brother);                    
                    brother = cst(brother,2);
                end
            end
        end    
        clear stateseqBack;
        
        % carry out the linear regressions
        for a = 1:alphabetsize
            tl(:,:,a) = (covMatInv * V_normalized * protoW(:,fTable(useIndices,a))')';
        end
    else  % case of using pseudoinverse for linear regression
        % normalize protoV to version V_normalized by dividing through
        % sqrt's of state sums
        if length(protoV(:,1)) > 1 
            normSums = sqrt(sum(protoV));
        else
            normSums = sqrt(protoV); % for 1-dim models
        end     
        V_normalized = protoV;
        for n = 1:nrUsed 
            if normSums(1,n) ~= 0
                V_normalized(:,n) = protoV(:,n) / normSums(1,n);  
            end
        end
        clear protoV;
        
        if verbose >= 1  % a diagnostic item, computed only in the more expensive modes      
            condition = cond(V_normalized);        
        end
        
        pinvV = pinv(V_normalized);
        clear V_normalized;
        
        protoW = zeros(dim,nodeNr + 1);  % same as in Wiener Hopf case        
        for n = nodePerms'
            if cst(n,9) == 0 % case of seeing a leave            
                protoW(:,n) = stateseqBack(l+1 - cst(n,7) + 1,:)';                
            else
                protoW(:,n) = protoW(:,cst(n,3)); % init with summed states of first child                
                brother = cst(cst(n,3),2); % find index of first brother of first child
                while brother ~= 0
                    protoW(:,n) = protoW(:,n) + protoW(:,brother);                    
                    brother = cst(brother,2);
                end
            end
        end    
        clear stateseqBack;
        for a = 1:alphabetsize
            W = protoW(:,fTable(useIndices,a));
            for n = 1:nrUsed   
                if normSums(1,n) ~= 0
                    W(:,n) = W(:,n)/ normSums(1,n); 
                end
            end  
            tl(:,:,a) = W * pinvV;
        end
        clear W;
    end
    clear protoW V_normalized;
    
    w0 = compW0(tl);   % compute invariant starting state
    
    
    % exit with NaN results if model has effective dim that is lower than dim
    [bigVtest, PiMattest, stateMattest] = bigVfromModel(tl,w0,evLength);
    effDim = rank(bigVtest, dimTol);
    if effDim < dim
        disp(sprintf(...
            'Warning: effective model dim in iteration %g smaller than requested dim. Exit from learnOOM_SuffixTree',...
            iteration));
        tlLearnt = NaN * zeros(dim, dim, alphabetsize);
        w0Learnt = NaN * zeros(dim,1);
        WeightPL = NaN * zeros(1,iter  );
        CondPL = NaN * zeros(1,iter  ); 
        LLTrainPL = NaN * zeros(1,iter  );
        LLTestPL =  NaN * zeros(1,iter  );
        return
    end
    
    if mkInter % make model interpretable w.r.t. characteristic event characterizer used
        % for initial model
        [tl, rho]  = makeInterpretable(tl,Cfwd, evLength);
        w0 = rho * w0;              
    end    
    
    [tlBack, w0Back] = reverseOOM(tl);  % compute reverse OOM    
    
    % compute reverse state sequence
    [stateseqBack, loglikBack, warnsB, setbacksB, actCounterB, changeCounterB] = ...
        compStateSeq_robust(tlBack, w0Back , trainDataRev, margin, setbackMargin, setbackLength);
    stateseqBack = reverseMat(stateseqBack,1);  % reorder into "forward" direction
    
    % collect "cheap" diagnostics
    LLTrainBPL(iteration) = loglikBack;
    modelWeight = sum(sum(sum(abs(tl),3)));
    WeightPL(iteration) = modelWeight;
    
    % if best loglik model is to be returned, assign current model to
    % output if its loglik on reverse sequence is best up to now
    if modelSelectionMode == 2 
        if loglikBack > loglikBest
            loglikBest = loglikBack;
            tlLearnt = tl; w0Learnt = w0;
        end
    end
    
    % diagnostic output 
    if verbose == 1        
        disp(sprintf('it %g: LLback %0.1f  W %0.1f  cond %0.1f  negsB %g   setbB %g   actB %g  changesB %g',...
            iteration, loglikBack, modelWeight, condition, warnsB, ...
            setbacksB, actCounterB, changeCounterB));
    elseif verbose == 2
        [stateseqFwd, loglik, warns, setbacks, actCounter, changeCounter] = ...
            compStateSeq_robust(tl, w0 , trainData, margin, setbackMargin, setbackLength);         
        [LLtest, warnPercentTest] = compAvLL_robust(tl, w0, testseqs, ...
            margin, setbackMargin, setbackLength );        
        LLTrainPL(iteration) = loglik;
        LLTestPL(iteration) = LLtest;
        CondPL(:,iteration) = condition;         
        disp(sprintf('it %g: LLtr %0.1f  LLback %0.1f  LLts %0.2f  W %0.1f  Cond %0.1f ',...
            iteration, loglik, loglikBack, LLtest, modelWeight,  condition));
        disp(sprintf('       negs %g   setb %g   act %g  changes %g    negsB %g   setbB %g   actB %g  changesB %g',...
            warns, setbacks, actCounter, changeCounter, warnsB, ...
            setbacksB, actCounterB, changeCounterB));
    end
    
    if  statePlotFlag  % plot fingerprints           
        figure(f1);
        subplot(plotGrid, plotGrid,iteration);
        plotOOMStateseq(stateseqFwd(2:501,:), trainData(1:500), alphabetsize); 
        if iteration == 1
            title('SuffixTree ES forward fingerprints');
        end
    end    
    
    tlOld = tl;
end

if printLegend & verbose == 1
    disp('********* Explanation of diagnostic categories *********');
    disp('it: iteration');
    disp('LLback: log likelihood of reverse model on train data');
    disp('W: summed absolute matrix entries of (all) operator matrices');
    disp('cond: condition number of argument matrix covMat [for WienerHopf] or V_normalized [w/o WienerHopf]');
    disp('negsB: Nr of unrecovered negative probs encountered when computing reverse state sequence');
    disp('       Should be 0, else model is dubiously unstable');
    disp('setbB: Nr of emergency recovery setback operations during computing reverse state sequence');
    disp('actB: Nr of "blurred" probabilities used during computing reverse state sequence');
    disp('changesB: Nr of recovered negative probs encountered when computing reverse state sequence');
end
if printLegend & verbose == 2
    disp('********* Explanation of diagnostic categories *********');
    disp('it: iteration');
    disp('LLtr: log likelihood of forward model on train data');
    disp('LLback: log likelihood of reverse model on train data');
    disp('LLts: average log likelihood of forward model on test data');
    disp('W: summed absolute matrix entries of (all) operator matrices');
    disp('cond: condition number of argument matrix covMat [for WienerHopf] or V_normalized [w/o WienerHopf]');
    disp('negs: Nr of unrecovered negative probs encountered when computing forward state sequence');
    disp('      Should be 0, else model is dubiously unstable');
    disp('setb: Nr of emergency recovery setback operations during computing forward state sequence');
    disp('act: Nr of "blurred" probabilities used during computing forward state sequence');
    disp('changes: Nr of recovered negative probs encountered when computing forward state sequence');
    disp('negsB: Nr of unrecovered negative probs encountered when computing reverse state sequence');
    disp('       Should be 0, else model is dubiously unstable');
    disp('setbB: Nr of emergency recovery setback operations during computing reverse state sequence');
    disp('actB: Nr of "blurred" probabilities used during computing reverse state sequence');
    disp('changesB: Nr of recovered negative probs encountered when computing reverse state sequence');
end

% if model of last iteration is to be returned, assign to output
if modelSelectionMode == 1
    tlLearnt = tl;
    w0Learnt = w0;
end

if verbose == 2
    figure(f2);clf;    
    subplot(4,1,1);
    plot(LLTestPL'); 
    title('LLtest'); 
    subplot(4,1,2);
    plot(LLTrainPL); 
    title('LLtrain');    
    subplot(4,1,3);
    plot(WeightPL');
    title('Weights');
    subplot(4,1,4);
    plot(CondPL); 
    title('condition numbers');    
end


if verbose > 0
    disp(' ************* finished learnOOM_SuffixTree ************ ');
end
