function [tlInt, w0Int, rho, C, charEvLength] = makeInterpretable_Auto(tl,w0)
% transforms an OOM specified by taulist tl into interpretable OOM with characteristic events
% computed automatically, using the heuristic algorithm
% findCharEventsFromBigV. For a description of the algorithm, check out Appendix I in H. Jaeger et al,
% "Learning Observable Operator Models via the ES Algorithm", in "New
% Directions in Statistical Signal Processing: from Systems to Brains"
% (eds. Simon Haykin, Jose Principe, Terry Sejnowski, John McWhirter), MIT Press 2005. 
%
% input arguments:
% tl: array of size dim x dim x alphabetsize, containing the operators of
%     an OOM
% w0: its starting state
%
% outputs:
% tlInt: array of size dim x dim x alphabetsize, containing the operators
%        of an interpretable OOM equivalent to tl
% w0Int: its starting state
% rho: the transformation matrix, size dim x dim, for transforming tl into
%      tlInt
% C: the characteristic event characterizer 
% charEvLength: the length of characteristic events used for the
% transformation
%
% V 1.0, May 2005
% (c) Herbert Jaeger
%
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.

alphabetsize = size(tl,3);
dim = size(tl,1);

% compute charEvLength
charEvLength = 1;
while alphabetsize^charEvLength < dim
    charEvLength = charEvLength + 1;
end
kappa = alphabetsize^charEvLength;

% compute the "big" matrix Vbig of size \kappa x \kappa, where \kappa = alphabetsize^charEvLength
% Vbig = (P(\bar{a}_j \bar{a}_i)), where the \bar{a}_i are
% all words of length charEvLength.

% first, compute the \kappa x dim matrix leftHalves with rows 
% \mathbf{1} * \tau_{\bar{a}_i} and the dim x \kappa matrix rightHalves
% with columns \tau_{\bar{a}_j} * w0
leftHalves = ones(kappa,dim);
rightHalves = w0 * ones(1,kappa);
for k = 1:kappa
    % compute k-th word "seq" of length charEvLength
    seq = zeros(1,charEvLength);
    j = k-1;
    for l = charEvLength : - 1 : 1        
        ind =  floor((j)/alphabetsize^(l-1));
        j = j - ind * alphabetsize^(l-1);        
        seq(1,charEvLength-l+1) = ind + 1;
    end  
    % fill in row in leftHalves
    for l = 1:charEvLength
        leftHalves(k,:) = leftHalves(k,:) * tl(:,:,seq(l));
    end
    % fill in column in rightHalves
    for l = 1:charEvLength
        rightHalves(:,k) = tl(:,:,seq(charEvLength - l + 1)) * rightHalves(:,k) ;
    end
end
% obtain bigV from these
bigV = leftHalves * rightHalves;

% find good characterizing event characterizer
[C, w0Int] = findCharEventsFromBigV(bigV, dim, 0);

% compute the dim x dim matrix rho with rows \mathbf{1} * \tau_{A_i}
rho = C * leftHalves;

if cond(rho) > 1e8
    disp('Warning: bad condition of rho in makeInterpretable_Auto, used hotfix. Model is not interpretable.');
    rho = rho + 0.01 * eye(dim);
    for d = 1:dim
        rho(:,d) = rho(:,d) / sum(rho(:,d));
    end
end

% carry out transformation
rhoInv = inv(rho);
tlInt = tl;
for i = 1:alphabetsize
    tlInt(:,:,i) = rho * tl(:,:,i) * rhoInv;
end

