function [] = plotOOMSequence_interpretable(tl, w0, C, evLength, sequence);
% generates a run of tl driven by sequence seq and plots states into currently active figure.
% The OOM may have any dimension d1 >= 3. C is a characterizer for an OOM of any (possibly
% other) dimension d2 >= 3. The length of characteristic sequences is evLength. 
% The original OOM states are first transformed
% into d2-dimensional states using C, then projected in 3-dim states for
% plotting.
% Up to 6 observables are plotted in different colours (b g r c k m). If
% there are more observables, all the remaining after the 6th are plotted
% in colour m. 
%
% input arguments:
% tl: size d1 x d1 x alphabetsize array containing observable operators.
% w0: starting state (column vector)
% C: characterizer matrix of size d2 x (alphabetsize^evLength)
% seq: sequence to be plotted.
%
% V 1.0, May 2005
% (c) Herbert Jaeger
%
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.

alphabetsize = size(tl,3);
dim = size(tl,1);
kappa = alphabetsize^evLength;
d2 = size(C,1);

if dim < 3
    disp('Warning: cannot plot states of dim < 3');
else
    % compute \pi matrix for projecting d1-states to
    % (alphabetsize^evLength)-dim states
    pi = ones(kappa, dim);
    for i = 1:kappa
        % generate i-th sequence in alphabetic enumeration
        seq = zeros(1,evLength);
        j = i-1;
        for k = evLength : - 1 : 1        
            ind =  floor((j)/alphabetsize^(k-1));
            j = j - ind * alphabetsize^(k-1);        
            seq(1,evLength-k+1) = ind + 1;
        end        
        for n = evLength:-1:1
            pi(i,:) = pi(i,:) * tl(:,:,seq(n));             
        end        
    end
    
    % compute (admittedly arbitrary) matrix for projecting d2-dimensional states to 3dim    
    projectionMat = zeros(3,d2);
    for d = 1:d2
        d3 = mod(d,3);
        if d3 == 0  d3 = 3; end
        projectionMat(d3,d) = 1;
    end
    % generate run of lenght numberOfPoints, collect states in 2 or 3 lists
    % according to observable producing the state
    statePlotLength = length(seq);
    modStatePL3D = zeros(statePlotLength,3,min([alphabetsize 6]));
    modStateCounters = ones(min([alphabetsize 6]));
    w = w0;    
    for n = 1:length(sequence)
        % update model       
        choice = sequence(n);
        w = tl(:,:,choice) * w;
        w = w / sum(w);
        w3 = projectionMat * C * pi * w ;
        modStatePL3D(modStateCounters(min([choice 6])), :, min([choice 6])) = w3;
        modStateCounters(min([choice 6])) = modStateCounters(min([choice 6])) + 1;         
    end
    
    % plot states
    colorList = 'bgrckm';
    hold on;
    for obsNr = 1:min([alphabetsize 6])
        plotStates3DColored(modStatePL3D(1:modStateCounters(obsNr)-1,:, obsNr),...
            obsNr, colorList);
    end
    % plot triangle
    plot([-0.5 * sqrt(2)  0.5 * sqrt(2) 0 -0.5 * sqrt(2)],...
        [-(1/3)* sqrt(1.5) -(1/3)* sqrt(1.5) (2/3)* sqrt(1.5) -(1/3)* sqrt(1.5)], 'k');    
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [] = plotStates3DColored(statelist3D, obsNr, colorList);
% plots a statelist into currently active figure
% statelist contains 3D states as rows
% obsNr gives index of a particular observation symbol. Different obsNr
% lead to different colors

l = length(statelist3D(:,1));
statelist2D = zeros(l,2);
for i = 1:l
    state3D = statelist3D(i,:);
    if state3D(1,1) + state3D(1,2) == 0
        statelist2D(i,:) =  [0 (2/3)* sqrt(1.5)];
    end
    statelist2D(i,:) = [sqrt(2) * (1 - state3D(1,3)) * ...
            (state3D(1,2)/(state3D(1,1) + state3D(1,2)) - 0.5), ...
            sqrt(1.5) * (state3D(1,3) - (1/3))];     
end
plot(statelist2D(:,1), statelist2D(:,2),  [colorList(obsNr), '.'] );

