function [] = plotOOMStateseq(statelist, obsList, alphabetsize)
% plots a sequence of OOM states in a "fingerprint" diagram, into the
% currently active figure window.
% The dim of the states must be at least 3. 
% If the dim is larger than 3, a projection on a 3dim subspace is plotted.
% Up to 6 observables are plotted in different colours (b g r c k m). If
% there are more observables, all the remaining afther the 6th are plotted
% in colour m.
%
% input arguments:
% statelist: contains states in rows.
% obsList:  contains the observables that produced states, 
%           is row vector of same length as size(statelist,1).
% alphabetsize: self-explanatory
%
% V 1.0, May 2005
% (c) Herbert Jaeger 
%
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.

alphabetsize = max(obsList);
modDim = length(statelist(1,:));
l = length(statelist(:,1));
if modDim < 3
    disp('Warning: cannot plot states of dim < 3');
else
    % compute matrix for projecting higher-dim states to 3dim    
    projectionMat = zeros(3,modDim);
    for d = 1:modDim
        d3 = mod(d,3);
        if d3 == 0  d3 = 3; end
        projectionMat(d3,d) = 1;
    end
    stateSeq3D = (projectionMat * statelist')';
    
    % reduce all observables with index > 6 to index 6
    obsListReduced = obsList;
    if alphabetsize > 6
        for n = 1:l
            if obsListReduced(1,n) > 6
                obsListReduced(1,n) = 6;
            end
        end
    end
        
    % disassemble into maximally 6 color-separated state plot lists
    modStatePL3D = zeros(l,3,min([alphabetsize 6]));
    modStateCounters = zeros(1,min([alphabetsize 6]));
    for n = 1:l
        index = obsListReduced(1,n);
        modStateCounters(1,index) = modStateCounters(1,index) + 1;
        modStatePL3D(modStateCounters(1,index),:,index) = stateSeq3D(n,:);
    end
    % plot triangle
    plot([-0.5 * sqrt(2)  0.5 * sqrt(2) 0 -0.5 * sqrt(2)],...
        [-(1/3)* sqrt(1.5) -(1/3)* sqrt(1.5) (2/3)* sqrt(1.5) -(1/3)* sqrt(1.5)], 'k');
    % plot states
    colorList = 'bgrckm';
    hold on;
    for obsNr = 1:min([alphabetsize 6])
        plotStates3DColored(modStatePL3D(1:modStateCounters(obsNr),:, obsNr),obsNr, colorList);
    end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [] = plotStates3DColored(statelist3D, obsNr, colorList);
% plots a statelist into currently active figure
% statelist contains 3D states as rows
% obsNr gives index of a particular observation symbol. Different obsNr
% lead to different colors

l = length(statelist3D(:,1));
statelist2D = zeros(l,2);
for i = 1:l
    state3D = statelist3D(i,:);
    if state3D(1,1) + state3D(1,2) == 0
        statelist2D(i,:) =  [0 (2/3)* sqrt(1.5)];
    end
    if abs(state3D(1,1) + state3D(1,2)) < 0.0000001
        statelist2D(i,:) = [NaN, NaN];
    else
        statelist2D(i,:) = [sqrt(2) * (1 - state3D(1,3)) * ...
                (state3D(1,2)/(state3D(1,1) + state3D(1,2)) - 0.5), ...
                sqrt(1.5) * (state3D(1,3) - (1/3))];  
    end
end
plot(statelist2D(:,1), statelist2D(:,2),  [colorList(obsNr), '.'] );

