function [tl_dimRed, w0_dimRed, delSV_MassRatio] = reduceDim(tl, w0, lss, redN, matDimRed, figNr)
% Reduces by redN the dimension of an OOM given by a taulist tl and its stationary
% state w0.
% 
% Inputs: 
% tl: array sized dim x dim x alphabetsize, containing the observable
%     operators
% w0: starting state (stationary)
% lss: length of the state sequence which is generated to assess the
%      invariant cone of the OOM. The larger, the more precisely are the
%      least important principal components determined - at the cost of
%      computing more states. 
% 
% redN: the number of dimensions that should be taken off. Must be an
%       integer smaller than dim.
% matDimRed: a binary flag. If 1, the produced output OOM tl_dimRed is a
%            set of operator matrices whose matrix dimension is reduced by
%            matDimRed. If 0, the produced output OOM tl_dimRed is a
%            set of operator matrices whose matrix dimension is the same as
%            the matrix dimension of the input model tl (but whose matrices
%            are degenerate, having rank dim - dimRed). Both options give
%            equivalent models, but option "0" is faster to compute.
% figNr: an integer. If 0, no graphical output is produced, if positive, 
%        fingerprint plots are sent to a figure with the Nr. figNr.
%
% Outputs: 
% tl_dimRed, w0_dimRed: the dimension-reduced OOM
% delSV_MassRatio: the fraction 
%             (weightsum of deleted SV's) / (weightsum of all SV'). 
%             Should be close to zero and far away from 1 for the low-dim
%             approximation tl_dimRed to be a reasonable approximation to
%             tl.
% 
% Created July 13, 2007, H. Jaeger
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.

dim = size(tl,1); alphabetsize = size(tl,3);
% compute charEvLength
charEvLength = 1;
while alphabetsize^charEvLength < dim
    charEvLength = charEvLength + 1;
end

%%%% transform to reverse characterized model    
[bigV, PiMat, stateMat] = bigVfromModel(tl, w0, charEvLength);
% compute new characterizer C, using  reverse model
[tlBack, w0Back] = reverseOOM(tl); 
kappa = alphabetsize ^ charEvLength;
Crev = zeros(dim, kappa);
for i = 1:kappa
    % generate i-th sequence seq in alphabetic enumeration
    seq = zeros(1,charEvLength);
    j = i-1;
    for k = charEvLength : - 1 : 1        
        ind =  floor((j)/alphabetsize^(k-1));
        j = j - ind * alphabetsize^(k-1);        
        seq(1,charEvLength-k+1) = ind + 1;
    end
    w = w0Back;
    for n = charEvLength:-1:1
        w = tlBack(:,:,seq(n)) * w; 
        sumw = sum(w);
        if sumw ~= 0
            w = w / sumw;
        end
    end
    Crev(:,i) = w;
end
[tl_reverseCharacterized, rho]  = makeInterpretable(tl, Crev, charEvLength);
w0_reverseCharacterized = compW0(tl_reverseCharacterized);


%%%% generate state sequence

margin = 0.001; setbackMargin = 0.1; setbackLength = 3;
[stateSeq, obsSeq, loglik, warns, setbacks, actCounter] = ...
    compStateObsSeq_robust(tl_reverseCharacterized, w0_reverseCharacterized, lss, ...
    margin, setbackMargin, setbackLength);

%%%% SVD magic: map states on the most significant lower-dim subspace
stateSeq = stateSeq';
[L, S, R] = svd(stateSeq);
SV_Mass_before = sum(sum(S));
for d = 1:redN
    S(dim + 1 - d, dim + 1 - d) = 0;
end
SV_Mass_after = sum(sum(S));
delSV_MassRatio = (SV_Mass_before - SV_Mass_after)/SV_Mass_before;
stateSeqDimreduced = L * S * R';
Statesums = sum(stateSeqDimreduced);
for n = 1:lss
    stateSeqDimreduced(:,n) = stateSeqDimreduced(:,n) / Statesums(1,n);
end

%%% reduce effective dim
projector = stateSeqDimreduced * pinv(stateSeq);
tl_effDimRed = tl_reverseCharacterized; 
for a = 1:alphabetsize
    tl_effDimRed(:,:,a) = projector * tl_effDimRed(:,:,a);
end
w0_effDimRed = compW0(tl_effDimRed);


if not(matDimRed) % case we are content with returning the rank-deficient model matrices
    % which still have the original matrix dimension
    tl_dimRed = tl_effDimRed;  w0_dimRed = w0_effDimRed;
else % case we want to return matrices of the reduced size
    [bigVforRed, bigWforRed, XX, YY] = bigVWfromModel(tl_effDimRed, w0_effDimRed, charEvLength);
    [charEventsForRed, w0_dimRed] = findCharEventsFromBigV(bigVforRed, dim-1, charEvLength);
    inverter = pinv(charEventsForRed * bigVforRed);
    tl_dimRed = zeros(dim-1, dim-1, alphabetsize);
    for a = 1:alphabetsize
        tl_dimRed(:,:,a) = charEventsForRed * bigWforRed(:,:,a) * inverter;
    end
    w0_dimRed = compW0(tl_dimRed);
end


if figNr > 0
    figure(figNr); clf; set(figNr, 'Position', [60         341        1170         282]);
    subplot(1,3,1);
    plotOOMStateseq(stateSeq', obsSeq', alphabetsize);
    title('Full states');
    subplot(1,3,2);
    plotOOMStateseq(stateSeqDimreduced', obsSeq', alphabetsize);
    title('Reduced states (from full)');
    subplot(1,3,3);    
    [stateseqPL, obsseqPL, loglik, warns, setbacks, actCounter] = ...
    compStateObsSeq_robust(tl_dimRed, w0_dimRed, lss, margin, setbackMargin, setbackLength);    
    plotOOMStateseq(stateseqPL, obsseqPL', alphabetsize);
    title('Reduced states (from new model)');
end
