function [cst, slink, childInfo] = sfxtree(seq)
% construct the compact suffix tree "cst" for a given string "seq"
%	usage:		[cst, slink, childInfo] = sfxtree(seq)
%	where
%	cst			compact suffix tree of seq, an [n][10] matrix;
%	slink		reduced suffix links, (exclude the chain of leaves);
%   childInfo:  structure with information about leaves underneath each node
%	seq			source string, a row vector of {1, 2, ..., k}'s,
%				we assume that length(seq) >= 2.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% An n-node suffix tree is represented by an [n][12] matrix cst, each row
% denotes a node, using the following members:
% _____________________________________________________________________
%	NO.		Name		Description
%	1		parent		index of its parent node
%	2		brother		index of its first younger brother
%	3		child		index of its first child
%	4		firstpos	start index of its incoming edge (include)
%	5		lastpos		end index of its incoming edge (exclude)
%	6		weight		length of suffix denoted by this node
%   7       position    for leaf nodes, the start pos of string that leads
%                       to leaf; 0 for internal nodes
%   8       depth       link distance of node from root, counting nr. of
%                       links
%   9       maxToLeaf   maximal link path length to leaf underneath node
%   10      leaveNr     nr of leaves below (and including) node
%   11      parentW     weight of parent link (0 for root)
%   12      startSym    first symbol of sequence annotating the parent link
%                       (0 for root)
% ---------------------------------------------------------------

% Other convention about the suffix tree:
%	1.	the root always occupies the first row;
%	2.	index 0 means a void pointer;
%	3.	root(parent) = cst(1,1) is used to save the number of nodes in
%		the suffix tree;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The chain of suffix links is represented by an [n][2] matrix, each row
% denotes a hidden node, by using the following variables:
% _____________________________________________________________________
%	NO.		Name		Description
%	1		anchor		above which explicit node the hiden node lies
%	2		position	distance from parent(anchor), be positive
% ---------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The leaf list childInfo is a structure with two fields. Field  childInfo(n).leaves is 
% a list of the leaves underneath or equal to node n. Field childInfo(n).leafPos
% is a list of the beginning positions of the paths of the leaves
% underneath (and including) node n. This is the same as the possible beginning
% positions of the substring that leads from the root to the node n.

% (c) Mingjie Zhao, Herbert Jaeger
% Created May 2005, Herbert Jaeger, based on suffix tree
% implementation by Mingjie Zhao 
% Revision Oct 2, 2007, Herbert Jaeger: use corrected (now linear-time)
% suffix-tree core algorithm version from Mingjie Zhao (was
% quadratic time in first version)

% Compute cst table for first 5 columns (= Mingjie's implementation
% of core suffix tree algorithm: the hard work part...) 

len = length(seq);

% field of "parent, next-brother, first-child, firstpos & lastpos"
[PA,BR,CH,FP,LP] = deal(1,2,3,4,5);
[ANC,POS] = deal(1,2);  % field of "anchor, position"

% initialize <cst> for the first character of <seq>
sl = length(seq); % sequence length
ope = sl + 1; % end char index of open edges
cst = repmat(-1, 2*sl+1, 5);  % <cst> has at most (2*sl+1) nodes
cst(1:2,:) = [2, 0, 2, 0, 0; ...  % number of nodes: 2
  1, 0, 0, 1, ope]; % <ope> means an open edge
slink = zeros(sl+1, 2); % reduced suffix links, (exclude the chain of leaves).
num_links = 0;
for ii = 2 : sl
  % to cope with the ii-th character <ch>
  ch = seq(ii);
  % to extend the open edges: do nothing!!!
  % to deal with the suffix links
  if num_links ~= 0
    n_link = num_links;
    num_links = 0;
    jj = 1;
    found = false;
    while (~found) & (jj <= n_link)
      anchor = slink(jj, ANC); % anchor of the jj-th hidden node
      if slink(jj,POS) == cst(anchor,LP)-cst(anchor,FP) % explicit hidden node
        % to search for <ch> below <anchor>
        nodeA = cst(anchor, CH);
        while (ch ~= seq(cst(nodeA, FP))) & (cst(nodeA, BR) ~= 0)
          nodeA = cst(nodeA, BR);
        end
        if ch == seq(cst(nodeA, FP)) % find <ch> above <nodeA>
          num_links = num_links + 1;
          slink(num_links, :) = [nodeA, 1];
          found = true;
        else  % not find <ch>
          % to create a new leaf below <anchor> after <nodeA>
          cst(1,1) = cst(1,1) + 1;
          cst(cst(1,1),:) = [anchor, ... % parent: anchor
            0, ... % younger brother: no
            0, ... % child: no
            ii, ... % firstpos = ii
            ope]; % open edge
          cst(nodeA, BR) = cst(1,1); % link the new node after <nodeA>
        end
      else  % the jj-th hiden node is implicit
        if ch == seq(cst(anchor,FP) + slink(jj,POS)) % find <ch>
          num_links = num_links + 1;
          slink(num_links, :) = [anchor, slink(jj,POS)+1];
          found = true;
        else  % not find <ch>
          % to insert a new node above <anchor> at position <slink(jj,POS)>
          % 1. create the new node
          cst(1,1) = cst(1,1) + 1;
          nodeA = cst(1,1);  % index of new created node
          parent = cst(anchor, PA);
          cst(nodeA,:) = [parent, ... % parent
            cst(anchor, BR), ... % younger brother
            anchor, ... % first child
            cst(anchor, FP), ... % firstpos
            cst(anchor, FP) + slink(jj, POS)];
          % 2. modify the suffix chain slink and sfxc
          if num_links ~= 0
            index = find(slink(1:num_links,ANC)==anchor);
            slink(index,POS) = slink(index,POS) - slink(jj,POS);
          end
          index = find(slink((jj+1):n_link,ANC) == anchor);
          slink(jj+index, ANC) = nodeA;
          % 3. kick <anchor> out from the family of parent
          cst(anchor, PA) = nodeA;
          cst(anchor, FP) = cst(nodeA, LP);
          if cst(parent, CH) == anchor % <anchor> is the first child
            cst(parent, CH) = nodeA;
          else
            % to find the youngest elder brother of <anchor>
            nodeB = cst(parent, CH);
            while cst(nodeB, BR) ~= anchor
              nodeB = cst(nodeB, BR);
            end
            cst(nodeB, BR) = nodeA;
          end
          % to create a new node under <nodeA> (after <anchor>)
          cst(1,1) = cst(1,1) + 1;
          cst(cst(1,1), :) = [nodeA, 0, 0, ii, ope];
          cst(anchor, BR) = cst(1,1);
        end
      end  % is an explicit/implicit node
      jj = jj + 1;
    end  % while ~found ch

    while jj <= n_link % there are more hidden nodes
      anchor = slink(jj, ANC); % anchor of the jj-th hidden node
      if slink(jj,POS) == cst(anchor,LP)-cst(anchor,FP) % explicit hidden node
        % to search for <ch> below <anchor> (there must be)
        nodeA = cst(anchor, CH);
        while ch ~= seq(cst(nodeA, FP))
          nodeA = cst(nodeA, BR);
        end
        num_links = num_links + 1;
        slink(num_links, :) = [nodeA, 1];
      else  % implicit hidden node
        num_links = num_links + 1;
        slink(num_links,:) = [anchor, slink(jj,POS) + 1];
      end
      jj = jj + 1;
    end
  end  % if num_links ~= 0

  % to cope with the root
  % 1. search for <ch> below the root
  nodeA = cst(1, CH);  % the first child of root
  while (ch ~= seq(cst(nodeA,FP))) & (cst(nodeA,BR) ~= 0)
    nodeA = cst(nodeA, BR);  % move to next node
  end
  if ch == seq(cst(nodeA,FP)) % find <ch>
    num_links = num_links + 1;
    slink(num_links, :) = [nodeA, 1];
  else  % not find <ch>
    % create a new node below the root after <nodeA>
    cst(1,1) = cst(1,1) + 1;
    cst(cst(1,1), :) = [1, 0, 0, ii, ope];
    cst(nodeA, BR) = cst(1,1); % link the new node after <nodeA>
  end
end

cst = cst(1:cst(1), :);
%slink = slink(1:num_links, :);
%cst = cst(1:num_nodes, :);
%cst(1,1) = num_nodes;

%%%%%%%%%%%%%%%%%%%%%%%%% compute columns 6 -- 12 of cst
n = cst(1,1);
cst = [cst zeros(n,7)];

% compute column 11 (weight of parent link)
for i = 2:n
    if cst(i,3) == 0  % is a leaf  
        cst(i,11) = len - cst(i,4) + 1;
    else
        minChildPos = len;
        currentChild = cst(i,3);
        while cst(currentChild,2) ~= 0
            minChildPos = min(minChildPos, cst(currentChild,4));
            currentChild = cst(currentChild,2);
        end
        cst(i,11) = minChildPos - cst(i,4);
    end
end

% compute depth column 8 of cst (link distance from root)

for i = 1:n
    currentNode = i;
    while currentNode ~= 1
        cst(i,8) = cst(i,8) + 1;
        currentNode = cst(currentNode, 1);
    end
end
linkDepth = max(cst(:,8));

% compute column 6 of cst (weight)
for d = 1:linkDepth
    for i = 1:n
        if cst(i,8) == d
            if d == 1
                cst(i,6) = cst(i,11);
            else
                cst(i,6) = cst(i,11) + cst(cst(i,1),6);
            end
        end
    end
end

% compute column 7 of cst

for i = 1:n
    if cst(i,3) == 0  % is a leaf        
        cst(i,7) =  len - cst(i,6) + 1;
    end
end



% compute maximal link path length of branch from node to leaf = column 9

for d = 1:linkDepth
    for i = 1:n         
        currentNode = i;        
        if cst(i,3) == 0  % is a leaf
            linkCounter = 1;
            while cst(cst(currentNode,1),9) < linkCounter & currentNode ~= 1
                currentNode = cst(currentNode,1); % go to parent
                cst(currentNode,9) = linkCounter;
                linkCounter = linkCounter + 1;
            end
        end       
    end
end

% compute symbol index of first symbol in sequence annotating incoming edge
% = column 12
for i = 2:n
    cst(i,12) = seq(cst(i,4));
end

% compute leaf list structure childInfo

childInfo(n) = struct('leaves',[]);
for d = 0:linkDepth
    for i = 1:n
        if cst(i,9) == d
            if d == 0 % we are seeing a leaf
                childInfo(i).leaves = [i];
                childInfo(i).leafPos = [cst(i,7)];
            else % collect leaf info from all children of current node
                % first collect info from first child
                childInfo(i).leaves = childInfo(cst(i,3)).leaves;
                childInfo(i).leafPos = childInfo(cst(i,3)).leafPos;
                % next collect info from brothers of first child
                currentChild = cst(i,3);
                while cst(currentChild,2) ~= 0
                    childInfo(i).leaves = [childInfo(i).leaves  childInfo(cst(currentChild,2)).leaves];
                    childInfo(i).leafPos = [childInfo(i).leafPos  childInfo(cst(currentChild,2)).leafPos];
                    currentChild = cst(currentChild,2);
                end
            end
        end
    end
end

% compute Nr of leaves below (and including) node

for i = 1:n
    cst(i,10) = length(childInfo(i).leaves);
end



