function [cst, slink, childInfo] = sfxtree(str)
% construct the compact suffix tree "cst" for a given string "str"
% using Ukkonen's online algorithm.
%	usage:		[cst, slink] = sfxtree(str)
%	where
%	cst			compact suffix tree of str, an [n][10] matrix;
%	slink		reduced suffix links, (exclude the chain of leaves);
%   childInfo:  structure with information about leaves underneath each node
%	str			source string, a row vector of {1, 2, ..., k}'s,
%				we assume that length(str) >= 2.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% An n-node suffix tree is represented by an [n][10] matrix cst, each row
% denotes a node, using the following members:
% _____________________________________________________________________
%	NO.		Name		Description
%	1		parent		index of its parent node
%	2		brother		index of its first younger brother
%	3		child		index of its first child
%	4		firstpos	start index of its incoming edge (include)
%	5		lastpos		end index of its incoming edge (exclude)
%	6		weight		length of suffix denoted by this node
%   7       position    for leaf nodes, the start pos of string that leads
%                       to leaf; 0 for internal nodes
%   8       depth       link distance of node from root, counting nr. of
%                       links
%   9       maxToLeaf   maximal link path length to leaf underneath node
%   10      leaveNr     nr of leaves below (and including) node
%   11      parentW     weight of parent link (0 for root)
%   12      startSym    first symbol of sequence annotating the parent link
%                       (0 for root)
% ---------------------------------------------------------------

% Other convention about the suffix tree:
%	1	the root always occupies the first row;
%	2	index 0 means a void pointer;
%	3	root(parent) = cst(1,1) is used to save the number of nodes in
%		the suffix tree;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The chain of suffix links is represented by an [n][2] matrix, each row
% denotes a hidden node, by using the following variables:
% _____________________________________________________________________
%	NO.		Name		Description
%	1		anchor		above which explicit node the hiden node lies
%	2		position	distance from parent(anchor), be positive
% ---------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The leaf list childInfo is a structure with two fields. Field  childInfo(n).leaves is 
% a list of the leaves underneath or equal to node n. Field childInfo(n).leafPos
% is a list of the beginning positions of the paths of the leaves
% underneath (and including) node n. This is the same as the possible beginning
% positions of the substring that leads from the root to the node n.

% (c) Mingjie Zhao, Herbert Jaeger 
% V 1.0, May 2005
%
% Copyright and limitation of liability: as specified in the document
% Copyright.txt that comes along with the OOM toolbox package from which 
% this file is taken.

% Compute cst table for first 5 columns 

PA = 1;  % parent
BR = 2;  % brother
CH = 3;  % child
FP = 4;  % firstpos
LP = 5;  % lastpos
ANC = 1;  % anchor
POS = 2;  % position

% initialize for the first character
len = length(str);
cst = repmat(-1, 2*len+1, 5);  % there are at most (2*len+1) nodes
cst(1:2, :) = [0, 0, 2, 0, 0; 1, 0, 0, 1, 2];
num_nodes = 2;
slink = zeros(len+1, 2);
num_links = 0;

for ii = 2 : length(str)
  ch = str(ii);  % to cope with the ii-th character

  % to extend the open edges
  idxleaves = find(cst(1:num_nodes, CH) == 0);
  cst(idxleaves, LP) = ii + 1;

  % to deal with the suffix links
  if num_links ~= 0
	num_links_bak = num_links;
	sfxc = slink(1:num_links, :);  % save the suffix links
	num_links = 0;
	jj = 1;
	found = false;
	while (~found) & (jj <= num_links_bak)
	  anchor = sfxc(jj, ANC);  % anchor of the jj-th hidden node
	  if sfxc(jj, POS) == cst(anchor, LP) - cst(anchor, FP)  % is an explicit node
		% to search ch below anchor
		nodeA = cst(anchor, CH);
		while (ch ~= str(cst(nodeA, FP))) & (cst(nodeA, BR) ~= 0)
		  nodeA = cst(nodeA, BR);
		end
		if ch == str(cst(nodeA, FP))  % find ch above nodeA
		  num_links = num_links + 1;
		  slink(num_links, :) = [nodeA, 1];
		  found = true;
		else  % not find ch 
		  % to create a new leaf below anchor (after nodeA)
		  num_nodes = num_nodes + 1;
		  cst(num_nodes, :) = [anchor, ... % parent: anchor
			  0, ... % younger brother: no
			  0, ... % child: no
			  ii, ... % firstpos = ii
			  ii + 1]; % lastpos = ii + 1
		  cst(nodeA, BR) = num_nodes;  % link the new node after nodeA
		end
	  else  % is an implicit node
		if ch == str(cst(anchor, FP) + sfxc(jj, POS))  % find ch
		  num_links = num_links + 1;
		  slink(num_links, :) = [anchor, sfxc(jj, POS) + 1];
		  found = true;
		else  % not find ch
		  % to insert a new node above anchor at position 'sfxc(jj, POS)'
		  % 1. create the new node
		  num_nodes = num_nodes + 1;
		  nodeA = num_nodes;  % index of the new created node
		  parent = cst(anchor, PA);
		  cst(num_nodes, :) = [parent, cst(anchor, BR), anchor, ...
			  cst(anchor, FP), cst(anchor, FP) + sfxc(jj, POS)];
		  % 2. modify the suffix chain slink and sfxc
		  if num_links ~= 0
			index = find(slink(1:num_links, ANC) == anchor);
			slink(index, POS) = slink(index, POS) - sfxc(jj, POS);
		  end
		  index = find((sfxc(:,ANC) == anchor) & (sfxc(:,POS) < sfxc(jj,POS)));
		  sfxc(index, ANC) = nodeA;
		  % 3. kick out anchor from the family of parent
		  cst(anchor, PA) = nodeA;
		  cst(anchor, FP) = cst(nodeA, LP);
		  if cst(parent, CH) == anchor  % anchor is the first child
			cst(parent, CH) = nodeA;
		  else
			% to find the elder brother of anchor
			nodeB = cst(parent, CH);
			while cst(nodeB, BR) ~= anchor
			  nodeB = cst(nodeB, BR);
			end
			cst(nodeB, BR) = nodeA;
		  end
		  % to create a new node under nodeA (after anchor)
		  num_nodes = num_nodes + 1;
		  cst(num_nodes, :) = [nodeA, 0, 0, ii, ii + 1];
		  cst(anchor, BR) = num_nodes;
		end
	  end  % is an explicit/implicit node
	  jj = jj + 1;
	end  % while ~found ch

	while jj <= num_links_bak  % there are more hidden nodes
	  anchor = sfxc(jj, ANC);  % anchor of the jj-th hidden node
	  if sfxc(jj, POS) == cst(anchor, LP) - cst(anchor, FP)  % is an explicit node
		% to search ch below anchor
		nodeA = cst(anchor, CH);
		while ch ~= str(cst(nodeA, FP))
		  nodeA = cst(nodeA, BR);
		end
		num_links = num_links + 1;
		slink(num_links, :) = [nodeA, 1];
	  else  % is an implicit node
		num_links = num_links + 1;
		slink(num_links, :) = [anchor, sfxc(jj, POS) + 1];
	  end  % is an explicit/implicit node
	  jj = jj + 1;
	end  % while jj <= size(sfxc, 1)
  end  % if ~isempty(slink)

  % to cope with root
  % 1. search ch below the root
  nodeA = cst(1, CH);  % the first child of root
  while (ch ~= str(cst(nodeA, FP))) & (cst(nodeA, BR) ~= 0)
	nodeA = cst(nodeA, BR);  % move to next node
  end
  if ch == str(cst(nodeA, FP)) % find ch
	num_links = num_links + 1;
	slink(num_links, :) = [nodeA, 1];
  else  % not find ch
	% to create a new node below the root (right after nodeA)
	num_nodes = num_nodes + 1;
	cst(num_nodes, :) = [1, 0, 0, ii, ii + 1];
	cst(nodeA, BR) = num_nodes;  % link the new node after nodeA
  end
end

slink = slink(1:num_links, :);
cst = cst(1:num_nodes, :);
cst(1,1) = num_nodes;

%%%%%%%%%%%%%%%%%%%%%%%%% compute columns 6 -- 11 of cst
n = cst(1,1);
cst = [cst zeros(n,7)];

% compute column 11 (weight of parent link)
for i = 2:n
    if cst(i,3) == 0  % is a leaf  
        cst(i,11) = len - cst(i,4) + 1;
    else
        minChildPos = len;
        currentChild = cst(i,3);
        while cst(currentChild,2) ~= 0
            minChildPos = min(minChildPos, cst(currentChild,4));
            currentChild = cst(currentChild,2);
        end
        cst(i,11) = minChildPos - cst(i,4);
    end
end

% compute depth column 8 of cst (link distance from root)

for i = 1:n
    currentNode = i;
    while currentNode ~= 1
        cst(i,8) = cst(i,8) + 1;
        currentNode = cst(currentNode, 1);
    end
end
linkDepth = max(cst(:,8));

% compute column 6 of cst (weight)
for d = 1:linkDepth
    for i = 1:n
        if cst(i,8) == d
            if d == 1
                cst(i,6) = cst(i,11);
            else
                cst(i,6) = cst(i,11) + cst(cst(i,1),6);
            end
        end
    end
end

% compute column 7 of cst

for i = 1:n
    if cst(i,3) == 0  % is a leaf        
        cst(i,7) =  len - cst(i,6) + 1;
    end
end



% compute maximal link path length of branch from node to leaf = column 9

for d = 1:linkDepth
    for i = 1:n         
        currentNode = i;        
        if cst(i,3) == 0  % is a leaf
            linkCounter = 1;
            while cst(cst(currentNode,1),9) < linkCounter & currentNode ~= 1
                currentNode = cst(currentNode,1); % go to parent
                cst(currentNode,9) = linkCounter;
                linkCounter = linkCounter + 1;
            end
        end       
    end
end

% compute symbol index of first symbol in sequence annotating incoming edge
% = column 12
for i = 2:n
    cst(i,12) = str(cst(i,4));
end

% compute leaf list structure childInfo

childInfo(n) = struct('leaves',[]);
for d = 0:linkDepth
    for i = 1:n
        if cst(i,9) == d
            if d == 0 % we are seeing a leaf
                childInfo(i).leaves = [i];
                childInfo(i).leafPos = [cst(i,7)];
            else % collect leaf info from all children of current node
                % first collect info from first child
                childInfo(i).leaves = childInfo(cst(i,3)).leaves;
                childInfo(i).leafPos = childInfo(cst(i,3)).leafPos;
                % next collect info from brothers of first child
                currentChild = cst(i,3);
                while cst(currentChild,2) ~= 0
                    childInfo(i).leaves = [childInfo(i).leaves  childInfo(cst(currentChild,2)).leaves];
                    childInfo(i).leafPos = [childInfo(i).leafPos  childInfo(cst(currentChild,2)).leafPos];
                    currentChild = cst(currentChild,2);
                end
            end
        end
    end
end

% compute Nr of leaves below (and including) node

for i = 1:n
    cst(i,10) = length(childInfo(i).leaves);
end



